# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Image encoding, decoding, and downsampling utility."""
import enum
import io
from typing import Optional

import cv2
import numpy as np

try:
  import PIL.Image

  _PIL_LOADED = True
except ImportError:
  _PIL_LOADED = False


# DICOM Transfer syntax define the encoding of the pixel data in an instance.
# https://www.dicomlibrary.com/dicom/transfer-syntax/
class DicomTransferSyntax(enum.Enum):
  JPEG_BASELINE = '1.2.840.10008.1.2.4.50'
  JPEG_2000_LOSSLESS = '1.2.840.10008.1.2.4.90'
  JPEG_2000 = '1.2.840.10008.1.2.4.91'


def can_decompress_dicom_transfer_syntax(transfer_syntax: str) -> bool:
  return transfer_syntax in (
      supported_syntax.value for supported_syntax in DicomTransferSyntax
  )


def decode_dicom_compressed_frame_bytes(frame: bytes) -> Optional[np.ndarray]:
  """Decode compressed frame bytes to DICOM BGR image.

  Args:
    frame: Raw image bytes (compressed blob).

  Returns:
    Decompressed image or None if decompression fails.
  """
  result = cv2.imdecode(np.frombuffer(frame, dtype=np.uint8), cv2.IMREAD_COLOR)
  if result is not None:
    cv2.cvtColor(result, cv2.COLOR_BGR2RGB, dst=result)
    return result
  if _PIL_LOADED:
    try:
      return np.asarray(PIL.Image.open(io.BytesIO(frame)))
    except PIL.UnidentifiedImageError:
      pass
  return None



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions for converting Magnification to PixelSpacing & vice versa.

These functions are separated from the PixelSpacing class so they can be more
easily removed once Magnification is fully removed.
"""
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import magnification as mag_lib
from ez_wsi_dicomweb import pixel_spacing


def magnification_to_pixel_spacing(
    magnification: mag_lib.Magnification,
) -> pixel_spacing.PixelSpacing:
  """Lightweight function that converts Magnification to PixelSpacing.

  Args:
    magnification: an instance of EZ WSI's magnification class to convert.

  Returns:
    A PixelSpacing instance created using the input magnification.

  Raises:
    MagnificationLevelNotFoundError if the provided magnification is unknown.
    InvalidMagnificationStringError if the string is not a non-zero positive
      floating point number.
  """
  if magnification.is_unknown:
    raise ez_wsi_errors.MagnificationLevelNotFoundError(
        'Provided magnification cannot be converted to PixelSpacing'
    )
  return pixel_spacing.PixelSpacing.FromMagnificationString(
      magnification.as_string
  )


def magnification_string_to_pixel_spacing(
    magnification: str,
) -> pixel_spacing.PixelSpacing:
  """Lightweight function that converts Magnification str to PixelSpacing."""
  return pixel_spacing.PixelSpacing.FromMagnificationString(magnification)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A class that represents a WSI slide magnification."""

from __future__ import annotations
import dataclasses
import enum
import math
from typing import Dict


@enum.unique
class MagnificationLevel(enum.Enum):
  """Enums of all supported magnification levels for DICOM/WSI slides."""

  UNKNOWN_MAGNIFICATION = 0
  M_100X = 1  # 100x
  M_80X = 2  # 80X
  M_40X = 3  # 40x
  M_20X = 4  # 20x
  M_10X = 5  # 10x
  M_5X = 6  # 5x
  M_5X_DIV_2 = 7  # 2.5x = 5x / 2
  M_5X_DIV_4 = 8  # 1.25x = 5x / 4
  M_5X_DIV_8 = 9  # 0.625x = 5x / 8
  M_5X_DIV_16 = 10  # 0.3125x = 5x / 16
  M_5X_DIV_32 = 11  # 0.15625 = 5x / 32
  M_5X_DIV_64 = 12  # 0.078125 = 5x / 64
  M_5X_DIV_128 = 13  # 0.0390625 = 5x / 128


@dataclasses.dataclass(frozen=True)
class MagnificationProperties:
  """Class for storing all properties associated with a magnification.

  Used to build a fast lookup table in
  _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES.

  Attributes:
    as_double: The floating point value of the magnification.
    as_string: The string label of the magnification.
    nominal_pixel_size: The nominal pixel size in mpp (micrometers per pixel).
  """

  as_double: float
  as_string: str
  nominal_pixel_size: float


# A dict that maps all magnification levels to their properties.
_MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES: Dict[
    MagnificationLevel, MagnificationProperties
] = {
    MagnificationLevel.UNKNOWN_MAGNIFICATION: MagnificationProperties(
        as_double=-1.0, as_string='unknown', nominal_pixel_size=0
    ),
    MagnificationLevel.M_100X: MagnificationProperties(
        as_double=100.0, as_string='100X', nominal_pixel_size=0.1
    ),
    MagnificationLevel.M_80X: MagnificationProperties(
        as_double=80.0, as_string='80X', nominal_pixel_size=0.125
    ),
    MagnificationLevel.M_40X: MagnificationProperties(
        as_double=40.0, as_string='40X', nominal_pixel_size=0.25
    ),
    MagnificationLevel.M_20X: MagnificationProperties(
        as_double=20.0, as_string='20X', nominal_pixel_size=0.5
    ),
    MagnificationLevel.M_10X: MagnificationProperties(
        as_double=10.0, as_string='10X', nominal_pixel_size=1.0
    ),
    MagnificationLevel.M_5X: MagnificationProperties(
        as_double=5.0, as_string='5X', nominal_pixel_size=2.0
    ),
    MagnificationLevel.M_5X_DIV_2: MagnificationProperties(
        as_double=2.5, as_string='2.5X', nominal_pixel_size=4.0
    ),
    MagnificationLevel.M_5X_DIV_4: MagnificationProperties(
        as_double=1.25, as_string='1.25X', nominal_pixel_size=8.0
    ),
    MagnificationLevel.M_5X_DIV_8: MagnificationProperties(
        as_double=0.625, as_string='0.625X', nominal_pixel_size=16
    ),
    MagnificationLevel.M_5X_DIV_16: MagnificationProperties(
        as_double=0.3125, as_string='0.3125X', nominal_pixel_size=32.0
    ),
    MagnificationLevel.M_5X_DIV_32: MagnificationProperties(
        as_double=0.15625, as_string='0.15625X', nominal_pixel_size=64.0
    ),
    MagnificationLevel.M_5X_DIV_64: MagnificationProperties(
        as_double=0.078125, as_string='0.078125X', nominal_pixel_size=128.0
    ),
    MagnificationLevel.M_5X_DIV_128: MagnificationProperties(
        as_double=0.0390625, as_string='0.0390625X', nominal_pixel_size=256.0
    ),
}

# The tolerance for matching a un-normalized pixel size value to a
# magnification level.
_MPP_SCALE_TOLERANCE = 0.2


class Magnification:
  """Represents a magnification level of a DICOM/WSI slide."""

  def __init__(self, magnification_level: MagnificationLevel):
    """Constructor.

    Args:
      magnification_level: The magnification.
    """
    self._magnification_level = magnification_level

  @classmethod
  def Unknown(cls):
    """Returns a Magnification object at UNKNOWN_MAGNIFICATION."""
    return Magnification(MagnificationLevel.UNKNOWN_MAGNIFICATION)

  @classmethod
  def FromString(cls, mag_str: str):
    """Returns a Magnification object given a string potentially representing a magnification.

    Args:
      mag_str: The string value of the magnfication to search for.
    """
    for (
        level,
        mag_level_props,
    ) in _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES.items():
      if mag_str == mag_level_props.as_string:
        return Magnification(level)
    return cls.Unknown()

  @classmethod
  def FromDouble(cls, mag_float: float):
    """Returns a Magnification by its floating value representation.

    Args:
      mag_float: The floating value of the magnfication to request.
    """
    for (
        mag_level,
        mag_level_props,
    ) in _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES.items():
      if math.isclose(mag_float, mag_level_props.as_double):
        return Magnification(mag_level)
    return cls.Unknown()

  @classmethod
  def FromPixelSize(cls, mpp: float):
    """Returns a Magnification by a pixel size.

    Args:
      mpp: The pixel size, in micrometers per pixel, of the magnification to
        request.
    """
    for (
        mag_level,
        mag_level_props,
    ) in _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES.items():
      if mag_level == MagnificationLevel.UNKNOWN_MAGNIFICATION:
        continue
      scale = mpp / mag_level_props.nominal_pixel_size
      if abs(scale - 1) < _MPP_SCALE_TOLERANCE:
        return Magnification(mag_level)
    return cls.Unknown()

  def __eq__(self, other: object) -> bool:
    if isinstance(other, Magnification):
      return self._magnification_level == other._magnification_level

    return False

  @property
  def nominal_pixel_size(self) -> float:
    """Returns the nominal pixel size of the magnification."""
    return _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES[
        self._magnification_level
    ].nominal_pixel_size

  @property
  def magnification_level(self) -> MagnificationLevel:
    """Returns the enum value of the magnification level."""
    return self._magnification_level

  @property
  def as_double(self) -> float:
    """Returns the floating value of the magnification level."""
    return _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES[
        self._magnification_level
    ].as_double

  @property
  def as_string(self) -> str:
    """Returns the string value of the magnification level."""
    return _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES[
        self._magnification_level
    ].as_string

  @property
  def is_unknown(self) -> bool:
    """Checks if the magnification is unknown."""
    return self._magnification_level == MagnificationLevel.UNKNOWN_MAGNIFICATION

  @property
  def next_higher_magnification(self) -> Magnification:
    """Creates a Magnification object containing next higher magnification level.

    Returns:
      An object with next higher magnification, or UNKNOWN magnification if this
      is the highest magnification level.
    """
    return self.FromPixelSize(
        _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES[
            self._magnification_level
        ].nominal_pixel_size
        / 2
    )

  @property
  def next_lower_magnification(self) -> Magnification:
    """Creates a Magnification object containing next lower magnification level.

    Returns:
      An object with the next lower magnification, or UNKNOWN magnification if
      this is the lowest magnification level.
    """
    return self.FromPixelSize(
        _MAGNIFICATION_LEVEL_TO_MAGNIFICATION_PROPERTIES[
            self._magnification_level
        ].nominal_pixel_size
        * 2
    )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A class that represents a WSI slide's pixel spacing."""
from __future__ import annotations

import math

from ez_wsi_dicomweb import ez_wsi_errors


# The tolerance (percentage difference) for difference between pixel spacings.
# EZ WSI expects practically square pixels.
_PIXEL_SPACING_DIFF_TOLERANCE = 0.01
# Magnification & Pixel Spacing are usually linearly related via a constant
# scaling factor of 0.01. he imaging scale factor can be directly derived via a
# single picture of a microscope calibration slide captured at a known
# magnification.
_SCALE_FACTOR = 0.01


class PixelSpacing:
  """Represents the pixel spacing of a DICOM/WSI slide.

  NOTE: By default the two provided spacings are averaged & used to compute a
    pixel spacing attribute that is used internally. This is because EZ WSI
    assumes square pixels.

  See also: https://dicom.innolitics.com/ciods/rt-dose/image-plane/00280030.
  """

  def __init__(
      self,
      row_spacing: float,
      column_spacing: float,
      scaling_factor: float = _SCALE_FACTOR,
      spacing_diff_tolerance: float = _PIXEL_SPACING_DIFF_TOLERANCE,
  ):
    """Constructor.

    Args:
      row_spacing: a number representing the pixel spacing measured vertically
        in mm/px.
      column_spacing: a number representing the pixel spacing measured
        horizontally in mm/px.
      scaling_factor: a number representing the scaling factor between a pixel
        spacing and a zoom level or magnification.
      spacing_diff_tolerance: The tolerance (percentage difference) for
        difference between row and column pixel spacings.

    Raises:
      NonSquarePixelError if the provided pixels are not roughly square.
    """

    if not math.isclose(
        row_spacing, column_spacing, rel_tol=spacing_diff_tolerance
    ):
      raise ez_wsi_errors.NonSquarePixelError(
          f'The provided row_spacing {row_spacing} and column_spacing'
          f' {column_spacing} are not square pixels.'
      )

    self._row_spacing = row_spacing
    self._column_spacing = column_spacing
    self._pixel_spacing = (row_spacing + column_spacing) / 2.0
    self._scaling_factor = scaling_factor
    self._spacing_diff_tolerance = spacing_diff_tolerance

  @classmethod
  def FromString(
      cls, pixel_spacing: str, scaling_factor: float = _SCALE_FACTOR
  ) -> PixelSpacing:
    """Returns a PixelSpacing object.

    Given a string representing a pixel spacing measured in in mm/px creates a
    PixelSpacing object.

    Args:
      pixel_spacing: The string value of the pixel spacing, e.g. a string
        representing a double measured in in mm/px.
      scaling_factor: a number representing the scaling factor between a pixel
        spacing and a zoom level or magnification.

    Returns:
      A new PixelSpacing object.
    """
    return PixelSpacing.FromDouble(
        pixel_spacing=float(pixel_spacing), scaling_factor=scaling_factor
    )

  @classmethod
  def FromMagnificationString(
      cls, magnification: str, scaling_factor: float = _SCALE_FACTOR
  ) -> PixelSpacing:
    """Returns a PixelSpacing object.

    Given a string representing a Magnificiation level creates a PixelSpacing
    object.

    Args:
      magnification: The string value of the magnification, e.g. a string
        representing a zoom level: 5, 5X, 10, 10X, 25X.
      scaling_factor: a number representing the scaling factor between a pixel
        spacing and a zoom level or magnification.

    Returns:
      A new PixelSpacing object.

    Raises:
      InvalidMagnificationStringError if the magnfication is not a float or <=0.
    """
    magnification = magnification.lower().rstrip('x')

    if not magnification.replace('.', '').isnumeric():
      raise ez_wsi_errors.InvalidMagnificationStringError(
          f'Provided magnification {magnification} is not a float.'
      )

    if float(magnification) <= 0:
      raise ez_wsi_errors.InvalidMagnificationStringError(
          f'Provided magnification {magnification} is <= 0. Magnifications must'
          ' be positive.'
      )

    pixel_spacing = scaling_factor / float(magnification)

    return PixelSpacing.FromDouble(
        pixel_spacing=pixel_spacing, scaling_factor=scaling_factor
    )

  @classmethod
  def FromDouble(
      cls, pixel_spacing: float, scaling_factor: float = _SCALE_FACTOR
  ) -> PixelSpacing:
    """Returns a PixelSpacing object.

    Given a pixel spacing float measured in mm/px creates a PixelSpacing object.

    Args:
      pixel_spacing: The avg value of the row & column pixel spacing in mm/px.
      scaling_factor: a number representing the scaling factor between a pixel
        spacing and a zoom level or magnification.

    Returns:
      A new PixelSpacing object.
    """
    return PixelSpacing(
        row_spacing=pixel_spacing,
        column_spacing=pixel_spacing,
        scaling_factor=scaling_factor,
    )

  def __hash__(self):
    return hash(
        tuple(
            (self.row_spacing_mm, self.column_spacing_mm, self.pixel_spacing_mm)
        )
    )

  def __eq__(self, other: object) -> bool:
    if isinstance(other, PixelSpacing):
      return math.isclose(
          self.row_spacing_mm,
          other.row_spacing_mm,
          rel_tol=self._spacing_diff_tolerance,
      ) and math.isclose(
          self.column_spacing_mm,
          other.column_spacing_mm,
          rel_tol=self._spacing_diff_tolerance,
      )

    return False

  @property
  def row_spacing_mm(self) -> float:
    return self._row_spacing

  @property
  def column_spacing_mm(self) -> float:
    return self._column_spacing

  @property
  def pixel_spacing_mm(self) -> float:
    return self._pixel_spacing

  @property
  def as_magnification_string(self) -> str:
    """Returns a string of the zoom level corresponding to the pixel spacing."""
    magnification = self._scaling_factor / self._pixel_spacing
    if magnification.is_integer():
      return f'{int(magnification)}X'
    else:
      return f'{magnification}X'



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.magnification."""

from typing import Union

from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import magnification


class MagnificationTest(parameterized.TestCase):

  def test_Unknown(self):
    mag = magnification.Magnification.Unknown()
    self.assertEqual(
        mag._magnification_level,
        magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
    )

  @parameterized.parameters(
      ('40X', magnification.MagnificationLevel.M_40X),
      ('0.0390625X', magnification.MagnificationLevel.M_5X_DIV_128),
      ('unknown', magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
      ('30X', magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
  )
  def test_FromString(
      self, str_value: str, expected_mag_level: magnification.MagnificationLevel
  ):
    self.assertEqual(
        expected_mag_level,
        magnification.Magnification.FromString(str_value)._magnification_level,
    )

  @parameterized.parameters(
      (40.0, magnification.MagnificationLevel.M_40X),
      (0.0390625, magnification.MagnificationLevel.M_5X_DIV_128),
      (-1, magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
      (30.0, magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
  )
  def test_FromDouble(
      self,
      float_value: float,
      expected_mag_level: magnification.MagnificationLevel,
  ):
    self.assertEqual(
        expected_mag_level,
        magnification.Magnification.FromDouble(
            float_value
        )._magnification_level,
    )

  @parameterized.parameters(
      (0.25, magnification.MagnificationLevel.M_40X),
      (256, magnification.MagnificationLevel.M_5X_DIV_128),
      (0, magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
      (0.19, magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
  )
  def test_FromPixelSize(
      self,
      pixel_size: float,
      expected_mag_level: magnification.MagnificationLevel,
  ):
    self.assertEqual(
        expected_mag_level,
        magnification.Magnification.FromPixelSize(
            pixel_size
        )._magnification_level,
    )

  @parameterized.parameters(
      (0.25, magnification.MagnificationLevel.M_40X),
      (256, magnification.MagnificationLevel.M_5X_DIV_128),
      (0, magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
  )
  def test_NominalPixelSize(
      self, pixel_size: float, mag_level: magnification.MagnificationLevel
  ):
    self.assertEqual(
        pixel_size, magnification.Magnification(mag_level).nominal_pixel_size
    )

  @parameterized.parameters(
      magnification.MagnificationLevel.M_40X,
      magnification.MagnificationLevel.M_5X_DIV_128,
      magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
  )
  def test_ToEnum(self, mag_level: magnification.MagnificationLevel):
    self.assertEqual(
        mag_level, magnification.Magnification(mag_level).magnification_level
    )

  @parameterized.parameters(
      (40.0, magnification.MagnificationLevel.M_40X),
      (0.0390625, magnification.MagnificationLevel.M_5X_DIV_128),
      (-1, magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
  )
  def test_ToDouble(
      self, float_value: float, mag_level: magnification.MagnificationLevel
  ):
    self.assertEqual(
        float_value, magnification.Magnification(mag_level).as_double
    )

  @parameterized.parameters(
      ('40X', magnification.MagnificationLevel.M_40X),
      ('0.0390625X', magnification.MagnificationLevel.M_5X_DIV_128),
      ('unknown', magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION),
  )
  def test_ToString(
      self, str_value: str, mag_level: magnification.MagnificationLevel
  ):
    self.assertEqual(
        str_value, magnification.Magnification(mag_level).as_string
    )

  @parameterized.parameters(
      (magnification.MagnificationLevel.M_40X, False),
      (magnification.MagnificationLevel.M_5X_DIV_128, False),
      (magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION, True),
  )
  def test_IsUnknown(
      self, mag_level: magnification.MagnificationLevel, is_unknown: bool
  ):
    self.assertEqual(
        is_unknown, magnification.Magnification(mag_level).is_unknown
    )

  @parameterized.parameters(
      (
          magnification.MagnificationLevel.M_40X,
          magnification.MagnificationLevel.M_80X,
      ),
      (
          magnification.MagnificationLevel.M_5X_DIV_128,
          magnification.MagnificationLevel.M_5X_DIV_64,
      ),
      (
          magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
          magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
      ),
      (
          magnification.MagnificationLevel.M_100X,
          magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
      ),
  )
  def test_NextHigher(
      self,
      mag_level: magnification.MagnificationLevel,
      next_level: magnification.MagnificationLevel,
  ):
    self.assertEqual(
        next_level,
        magnification.Magnification(
            mag_level
        ).next_higher_magnification.magnification_level,
    )

  @parameterized.parameters(
      (
          magnification.MagnificationLevel.M_40X,
          magnification.MagnificationLevel.M_20X,
      ),
      (
          magnification.MagnificationLevel.M_5X_DIV_64,
          magnification.MagnificationLevel.M_5X_DIV_128,
      ),
      (
          magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
          magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
      ),
      (
          magnification.MagnificationLevel.M_5X_DIV_128,
          magnification.MagnificationLevel.UNKNOWN_MAGNIFICATION,
      ),
  )
  def test_next_lower_magnification_level(
      self,
      mag_level: magnification.MagnificationLevel,
      next_level: magnification.MagnificationLevel,
  ):
    self.assertEqual(
        next_level,
        magnification.Magnification(
            mag_level
        ).next_lower_magnification.magnification_level,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='equal',
          mag_1=magnification.Magnification.FromString('40X'),
          mag_2=magnification.Magnification.FromString('40X'),
          expected_result=True,
      ),
      dict(
          testcase_name='not_equal',
          mag_1=magnification.Magnification.FromString('20X'),
          mag_2=magnification.Magnification.FromString('40X'),
          expected_result=False,
      ),
      dict(
          testcase_name='unknown_from_string_not_equal',
          mag_1=magnification.Magnification.FromString('blah'),
          mag_2=magnification.Magnification.FromString('40X'),
          expected_result=False,
      ),
      dict(
          testcase_name='from_double_equal',
          mag_1=magnification.Magnification.FromDouble(40),
          mag_2=magnification.Magnification.FromString('40X'),
          expected_result=True,
      ),
      dict(
          testcase_name='from_nominal_pixel_size_equal',
          mag_1=magnification.Magnification.FromPixelSize(0.25),
          mag_2=magnification.Magnification.FromString('40X'),
          expected_result=True,
      ),
      dict(
          testcase_name='different_types_not_equal',
          mag_1='Potato',
          mag_2=magnification.Magnification.FromString('40X'),
          expected_result=False,
      ),
  )
  def test_magnfications_equal(
      self,
      mag_1: Union[magnification.Magnification, str],
      mag_2: magnification.Magnification,
      expected_result: bool,
  ):
    self.assertEqual(mag_1 == mag_2, expected_result)
    self.assertEqual(mag_2 == mag_1, expected_result)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""An abstraction to sample patches in a slide at a specified maginfication."""

import dataclasses
from typing import Iterator, Optional

import cv2

from ez_wsi_dicomweb import dicom_slide
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import magnification as mag_lib
import numpy as np

TISSUE_MASK_MAGNIFICATION = mag_lib.Magnification.FromDouble(1.250)


@dataclasses.dataclass(frozen=True)
class StrideCoordinate:
  """Represents the coordinates of a stride using x, y coordinates.

  See PatchGenerator for usage examples.

  Attributes:
    y_strides: The number of strides along the y axis
    x_strides: The number of strides along the x axis
  """

  y_strides: int
  x_strides: int


class PatchGenerator:
  """A generator for patch sampling in a slide at specified magnification.

  The sampling process moves a sliding window of size patch_size with
  steps of size stride_size. The first sampled patch is centered at
  (-stride_size / 2, -stride_size / 2). We use (y_strides, x_strides) to denote
  how many strides we have taken along the y and x axis. Given a specified
  (y_strides, x_strides), a sampled patch can be uniquely identified and will
  be centered at (-stride_size / 2 + x_strides * stride_size,
  -stride_size / 2 + y_strides * stride_size).

  Utility functions are also provided to translate between strides and
  patch coordinates.
  """

  def __init__(
      self,
      slide: dicom_slide.DicomSlide,
      magnification: mag_lib.Magnification,
      stride_size: int,
      patch_size: int,
      max_luminance: float = 0.8,
      normalized_tissue_mask: Optional[np.ndarray] = None,
      tissue_mask_magnification: mag_lib.Magnification = TISSUE_MASK_MAGNIFICATION,
  ):
    """Constructor.

    Args:
      slide: The slide to generate on.
      magnification: The magnification level that we sample patches on.
      stride_size: Stride size at the magnification that we sample patches.
      patch_size: Patch size at the magnification that we sample patches.
      max_luminance: Regions with luminance (grayscale) > this threshold are to
        be considered non-tissue background, and will be discarded in the patch
        sampling.
      normalized_tissue_mask: If provided, will be cached as the tissue mask.
      tissue_mask_magnification: Used to determine where tissue is present if
        provided. If the provided DICOMSlide does not contain the provided
        magnification, patch generation will fail. This parameter must be
        overidden if the default magnification is not avaliable.
    """
    self.slide = slide
    self.magnification = magnification
    self.stride_size = stride_size
    self.patch_size = patch_size
    self.max_luminance = max_luminance
    self._normalized_tissue_mask_cache = normalized_tissue_mask
    self._tissue_mask_magnification = tissue_mask_magnification

  def __iter__(self) -> Iterator[dicom_slide.Patch]:
    """Given tissue mask, emits patches.

    Yields:
      A sampled patch.

    Raises:
      DicomImageMissingRegionError if there is no region with the
      required luminance values.
    """
    y_coords, x_coords = np.where(self._normalized_tissue_mask())
    for y, x in zip(y_coords, x_coords):
      yield self.patch_at_stride(StrideCoordinate(y, x))

  def _normalized_tissue_mask(self):
    """An image mask that is likely to contain tissues.

    It is normalized/down-scaled by the specified stride size, i.e.
    normalized_tissue_mask size = original size / stride.

    Returns:
      A 2D ndarray binary (0, 1) image denoting tissue mask.

    Raises:
      DicomImageMissingRegionError if there is no region with the
      required luminance values.
    """
    if self._normalized_tissue_mask_cache is not None:
      return self._normalized_tissue_mask_cache

    rgb_at_tissue_mag = self.slide.get_image(
        self._tissue_mask_magnification
    ).image_bytes()
    # luminance values
    if np.issubdtype(rgb_at_tissue_mag.dtype, np.floating):
      if rgb_at_tissue_mag.max() <= 1.0:
        rgb_at_tissue_mag *= 255.0
      rgb_at_tissue_mag = rgb_at_tissue_mag.astype(np.uint8)
    gray_tissue_mask = cv2.cvtColor(rgb_at_tissue_mag, cv2.COLOR_RGB2GRAY)
    tissue_mag_stride_size = (
        self.magnification.as_double / self._tissue_mask_magnification.as_double
    )
    scaling_factor = self.stride_size / tissue_mag_stride_size
    new_dimensions = (
        max(int(gray_tissue_mask.shape[1] // scaling_factor), 1),
        max(int(gray_tissue_mask.shape[0] // scaling_factor), 1),
    )
    gray_image_at_output_res = cv2.resize(
        gray_tissue_mask, new_dimensions, interpolation=cv2.INTER_AREA
    )
    self._normalized_tissue_mask_cache = (gray_image_at_output_res > 0) & (
        gray_image_at_output_res <= int(255.0 * self.max_luminance)
    )
    if np.all(~self._normalized_tissue_mask_cache):  # pylint: disable=invalid-unary-operand-type
      raise ez_wsi_errors.DicomImageMissingRegionError(
          f"Slide with study_uid {self.slide.path.study_uid} "
          f"series_uid {self.slide.path.series_uid} has no region with "
          "luminance value less than threshold "
          f"{self.max_luminance}."
      )
    return self._normalized_tissue_mask_cache

  def total_num_patches(self) -> int:
    """Total number of tissue patches in this slide."""
    y, _ = np.where(self._normalized_tissue_mask())
    return len(y)

  def total_strides(self) -> StrideCoordinate:
    y, x = self._normalized_tissue_mask().shape
    return StrideCoordinate(y_strides=y, x_strides=x)

  def strides_to_patch_bounds(
      self, strides: StrideCoordinate
  ) -> dicom_slide.PatchBounds:
    """Converts number of strides to patch coordinates.

    Args:
      strides: Number of strides taken on x and y axis.

    Returns:
      Patch Bounds corresponding to the strides.
    """
    center_y = self.stride_size * strides.y_strides + self.stride_size // 2
    center_x = self.stride_size * strides.x_strides + self.stride_size // 2
    top = center_y - self.patch_size // 2
    left = center_x - self.patch_size // 2

    return dicom_slide.PatchBounds(
        x_origin=left,
        y_origin=top,
        width=self.patch_size,
        height=self.patch_size,
    )

  def patch_bounds_to_strides(
      self, patch_bounds: dicom_slide.PatchBounds
  ) -> StrideCoordinate:
    """Converts patch bounds to number of strides in tissue mask.

    Args:
      patch_bounds: Describes the patch's coordinates bounding box.

    Returns:
      Corresponding number of strides.
    """
    half_stride = self.stride_size // 2
    half_patch_size = self.patch_size // 2
    center_y = patch_bounds.y_origin + half_patch_size - half_stride
    center_x = patch_bounds.x_origin + half_patch_size - half_stride
    y = center_y // self.stride_size
    x = center_x // self.stride_size
    return StrideCoordinate(y_strides=y, x_strides=x)

  def patch_at_stride(self, strides: StrideCoordinate) -> dicom_slide.Patch:
    """Converts tissue mask coordinates to patch coordinates.

    Args:
      strides: Number of strides taken from x and y axis.

    Returns:
      Corresponding Patch.
    """
    patch_bounds = self.strides_to_patch_bounds(strides)

    return self.slide.get_patch(
        self.magnification,
        x=patch_bounds.x_origin,
        y=patch_bounds.y_origin,
        width=patch_bounds.width,
        height=patch_bounds.height,
    )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""An interface for accessing a Pathology slide stored in a DICOMStore.

Whole-Slide Images (WSI) (https://dicom.nema.org/dicom/dicomwsi) in digital
pathology are scans of glass slides typically created at a magnification of
20x or 40x using microscopes. Pixels typically represent between
0.275 µm - 0.5 µm of a slide but the exact size depends on the scanner camera
resolution. With typical dimensions between of 15mm x 15mm for tissue slides,
WSIs are often about 50,000 x 50,000 pixels for a 20x magnification or
100,000 x 100,000 pixels for a 40x magnification. The images are usually stored
as image pyramid to allow for fast access to different magnification levels.

The class layout/hierarchy is as follows:
DicomSlide: a single (whole) slide image typically represented as
      an image pyramid along with slide level metadata.
   |--> DicomImage: an image at a specific magnification in the image pyramid of
        a slide along with associated metadata.
     |--> DicomImagePatch: a rectangular patch/tile of an Image
"""
from __future__ import annotations

import dataclasses
import heapq
import math
from typing import Any, Generator, Iterator, List, Mapping, MutableMapping, Optional, Tuple

import cachetools
from ez_wsi_dicomweb import abstract_slide_frame_cache
from ez_wsi_dicomweb import dicom_frame_decoder
from ez_wsi_dicomweb import dicom_web_interface
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import magnification as magnification_module
from ez_wsi_dicomweb import slide_level_map
import numpy as np

from hcls_imaging_ml_toolkit import dicom_path
from hcls_imaging_ml_toolkit import tags


@dataclasses.dataclass(frozen=True)
class _PatchIntersection:
  x_origin: int  # The upper leftmost x coordinate of the patch intersection.
  y_origin: int  # The upper leftmost y coordinate of the patch intersection.
  width: int
  height: int


@dataclasses.dataclass(frozen=True)
class Frame:
  """A frame in a DICOM, in pixel unit."""

  x_origin: int  # The upper leftmost x coordinate of the patch intersection.
  y_origin: int  # The upper leftmost y coordinate of the patch intersection.
  width: int
  height: int
  image_np: np.ndarray


# DICOM Transfer Syntax's which do not require transcoding and are natively
# compatible.
_SUPPORTED_CLIENT_SIDE_DECODING_RAW_TRANSFER_SYNTAXS = (
    '1.2.840.10008.1.2.1',
    '1.2.840.10008.1.2',
)


def is_client_side_pixel_decoding_supported(transfer_syntax: str) -> bool:
  """Returns True if cache supports client side decoding of pixel encoding.

  Args:
    transfer_syntax: DICOM transfer syntax uid.

  Returns:
    True if cache supports operation on instances with encoding.
  """
  return (
      dicom_frame_decoder.can_decompress_dicom_transfer_syntax(transfer_syntax)
      or transfer_syntax in _SUPPORTED_CLIENT_SIDE_DECODING_RAW_TRANSFER_SYNTAXS
  )


class Image:
  """Represents an image at a specific magnification in a DicomSlide."""

  def __init__(
      self, magnification: magnification_module.Magnification, slide: DicomSlide
  ):
    """Constructor for DicomImage.

    Args:
      magnification: The magnification level the image belongs to.
      slide: The parent slide this image belongs to.

    Raises:
      ValueError if the requested magnification is not valid.
    """
    self.magnification = magnification
    self._slide = slide

    level_at_mag = slide.get_level_by_magnification(magnification)
    if not level_at_mag:
      raise ez_wsi_errors.MagnificationLevelNotFoundError(
          f'No level found at magnification {magnification}'
      )

    self.width = level_at_mag.width
    self.height = level_at_mag.height

  @property
  def slide(self):
    return self._slide

  def image_bytes(self) -> np.ndarray:
    """Loads the pixel bytes of the DICOM Image.

    Returns:
      Numpy array representing the DICOM Image.
    """
    # Internally reuses the Patch implementation for bytes fetching.
    # An image can be represented as a giant patch starting from (0, 0)
    # and spans the whole slide.
    return self._slide.get_patch(
        magnification=self.magnification,
        x=0,
        y=0,
        width=self.width,
        height=self.height,
    ).image_bytes()


@dataclasses.dataclass(frozen=True)
class PatchBounds:
  """A bounding rectangle of a patch, in pixel units."""

  x_origin: int  # The upper leftmost x coordinate of the patch intersection.
  y_origin: int  # The upper leftmost y coordinate of the patch intersection.
  width: int
  height: int


class Patch:
  """A rectangular patch/tile/view of an Image at a specific magnification.

  A Patch's data is composed from its overlap with one or more DICOM Frames.
  """

  def __init__(
      self,
      magnification: magnification_module.Magnification,
      x: int,
      y: int,
      width: int,
      height: int,
      slide: DicomSlide = None,
  ):
    """Constructor.

    Args:
      magnification: The magnification level the patch belongs to. It defines
        the pixel space for the coordinates (x, y, width, height).
      x: The X coordinate of the starting point (upper-left corner) of the
        patch.
      y: The Y coordinate of the starting point (upper-left corner) of the
        patch.
      width: The width of the patch.
      height: The height of the patch.
      slide: The parent DICOM Slide this patch belongs to.
    """
    self.magnification = magnification
    self.x = x
    self.y = y
    self.width = width
    self.height = height
    self._slide = slide

  def __eq__(self, other: Any) -> bool:
    if not isinstance(other, Patch):
      return False

    return (
        self.magnification == other.magnification
        and self.x == other.x
        and self.y == other.y
        and self.width == other.width
        and self.height == other.height
        and self._slide == other._slide
    )

  @property
  def id(self) -> str:
    if not self.slide or not self.slide.accession_number:
      return (
          f'M_{self.magnification.as_string}:'
          f'{self.width:06d}x{self.height:06d}{self.x:+07d}{self.y:+07d}'
      )
    # Be consistent with internal id format.
    # "%s:%06dx%06d%+07d%+07d", image_id, width, height, left, top
    return (
        f'{self.slide.accession_number}:M_{self.magnification.as_string}:'
        f'{self.width:06d}x{self.height:06d}{self.x:+07d}{self.y:+07d}'
    )

  @property
  def slide(self):
    return self._slide

  @property
  def patch_bounds(self):
    return PatchBounds(
        x_origin=self.x, y_origin=self.y, width=self.width, height=self.height
    )

  def _get_intersection(
      self,
      frame_x_cord: int,
      frame_y_cord: int,
      frame_width: int,
      frame_height: int,
  ) -> _PatchIntersection:
    """Returns intersection from the source patch and rectangular coordinate.

    Args:
      frame_x_cord: Source frame x coordinate (upper left).
      frame_y_cord: Source frame y coordinate (upper left).
      frame_width: Source frame width
      frame_height: Source frame height

    Returns:
      Intersection between patches and rectangular coordinate.

    Raises:
      PatchIntersectionNotFoundError if there is no overlap between the source
      and the destination patches.
    """
    x = max(frame_x_cord, self.x)
    y = max(frame_y_cord, self.y)
    width = min(frame_x_cord + frame_width - x, self.x + self.width - x)
    height = min(frame_y_cord + frame_height - y, self.y + self.height - y)
    if width <= 0 or height <= 0:
      raise ez_wsi_errors.PatchIntersectionNotFoundError(
          'There is no overlap region between the source and the destination'
          ' patches.'
      )
    return _PatchIntersection(
        x_origin=x, y_origin=y, width=width, height=height
    )

  def _copy_overlapped_region(
      self, src_frame: Frame, dst_image_np: np.ndarray
  ) -> Tuple[int, int]:
    """Copies the overlapped region from the source patch to the dest nd.array.

    Args:
      src_frame: the source frame to be copied from.
      dst_image_np: the destination np to be copied to.

    Returns:
      The width and height of the overlapped region gets copied.

    Raises:
      PatchIntersectionNotFoundError if there is no overlap between the source
      and the destination patches.
    """
    intersection = self._get_intersection(
        src_frame.x_origin,
        src_frame.y_origin,
        src_frame.width,
        src_frame.height,
    )
    x = intersection.x_origin
    y = intersection.y_origin
    width = intersection.width
    height = intersection.height
    # pylint: disable=protected-access
    _copy_ndarray(
        src_frame.image_np,
        x - src_frame.x_origin,
        y - src_frame.y_origin,
        width,
        height,
        dst_image_np,
        x - self.x,
        y - self.y,
    )
    # pylint: enable=protected-access
    return width, height

  def frame_number(self) -> Generator[int, None, None]:
    """Yields slide level frame numbers required to render patch.

    Frame numbering starts at 0.
    """
    if self._slide is None:
      raise ez_wsi_errors.DicomSlideMissingError(
          'Unable to get image pixels. Parent slide is None.'
      )
    level = self._slide.get_level_by_magnification(self.magnification)
    if not level:
      raise ez_wsi_errors.MagnificationLevelNotFoundError(
          f'No level found at magnification {self.magnification}'
      )

    y = self.y
    x = self.x
    width = self.width
    height = self.height
    cy = y
    frame_width = level.frame_width
    frame_height = level.frame_height
    cached_instance = None
    last_instance_frame_index = -1
    while cy < y + height:
      cx = x
      region_height = 0
      while cx < x + width:
        try:
          frame_number = level.get_frame_number_by_point(cx, cy)
          if (
              cached_instance is None
              or frame_number - cached_instance.frame_offset
              >= cached_instance.frame_count
          ):
            cached_instance = level.get_instance_by_frame(frame_number)
            last_instance_frame_index = -1
            if cached_instance is None:
              raise ez_wsi_errors.FrameNumberOutofBoundsError()
          instance_frame_index = cached_instance.frame_index_from_frame_number(
              frame_number
          )
          if instance_frame_index > last_instance_frame_index:
            yield instance_frame_index + cached_instance.frame_offset - 1
          last_instance_frame_index = instance_frame_index
          pos_x, pos_y = level.get_frame_position(frame_number)
          intersection = self._get_intersection(
              pos_x, pos_y, frame_width, frame_height
          )
          region_width = intersection.width
          region_height = intersection.height
        except ez_wsi_errors.EZWsiError:
          # No frame found at (cx, cy), move 1 pixel in both X, and Y direction.
          region_width = 1
          region_height = max(1, region_height)
        cx += region_width
      cy += region_height

  def image_bytes(self) -> np.ndarray:
    """Returns the patch's image bytes.

    Returns:
      Numpy array type image.

    Raises:
      DicomSlideMissingError if slide used to create self is None.
      PatchOutOfBoundsError if the patch is not within the bounds of the DICOM
      image.
    """
    if self._slide is None:
      raise ez_wsi_errors.DicomSlideMissingError(
          'Unable to get image pixels. Parent slide is None.'
      )
    level = self._slide.get_level_by_magnification(self.magnification)
    if not level:
      raise ez_wsi_errors.MagnificationLevelNotFoundError(
          f'No level found at magnification {self.magnification}'
      )

    image_bytes = np.zeros(
        (self.height, self.width, level.samples_per_pixel),
        self.slide.pixel_format,
    )
    pixel_copied = False
    # Copies image pixels from all overlapped frames.
    for frame_number in self.frame_number():
      frame = self.slide.get_frame(self.magnification, frame_number)
      if frame is None:
        continue
      try:
        self._copy_overlapped_region(frame, image_bytes)
        pixel_copied = True
      except ez_wsi_errors.EZWsiError:
        continue
    if not pixel_copied:
      raise ez_wsi_errors.SectionOutOfImageBoundsError(
          'The requested patch is out of scope of the image.'
      )
    return image_bytes


def _get_native_level(
    slm: slide_level_map.SlideLevelMap,
) -> slide_level_map.Level:
  """Gets the native level from a SlideLevelMap.

  The native level of a slide has the lowest level index.

  Args:
    slm: The source SlideLevelMap to get the native level from.

  Returns:
    The level that has the lowest index in the slide.

  Raises:
    MagnificationLevelNotFoundError if the native level does not exist.
  """
  level = slm.get_level(slm.level_index_min)
  if level is not None:
    return level
  else:
    raise ez_wsi_errors.MagnificationLevelNotFoundError(
        'The native level is missing.'
    )


def _get_magnification(
    level: slide_level_map.Level,
) -> magnification_module.Magnification:
  """Gets the magnification corresponding to the input level.

  Args:
    level: The level to find the magnification of.

  Returns:
    The magnification corresponding to the input level.
  """
  pixel_spacing_um = (
      max(level.pixel_spacing_x_mm, level.pixel_spacing_y_mm) * 1000
  )
  return magnification_module.Magnification.FromPixelSize(pixel_spacing_um)


def _get_pixel_format(level: slide_level_map.Level) -> np.dtype:
  """Gets the pixel format of the provided level.

  Args:
    level: The level to get the pixel format for.

  Returns:
    The pixel format of the level as numpy.dtype.

  Raises:
    UnsupportedPixelFormatError if pixel format is not supported.
  """
  bytes_per_sample = math.ceil(level.bits_allocated / 8)
  if bytes_per_sample == 1:
    return np.uint8  # pytype: disable=bad-return-type  # numpy-scalars
  else:
    raise ez_wsi_errors.UnsupportedPixelFormatError(
        f'Pixel format not supported. BITS_ALLOCATED = {level.bits_allocated}'
    )


def _copy_ndarray(
    src_array: np.ndarray,
    src_x: int,
    src_y: int,
    width: int,
    height: int,
    dst_array: np.ndarray,
    dst_x: int,
    dst_y: int,
) -> None:
  """Copies a sub-array of an ndarray to another ndarray.

  Args:
    src_array: The source ndarray to copy the sub-array from.
    src_x: The X coordinate of the starting point to copy from, in the source
      array.
    src_y: The Y coordinate of the starting point to copy from, in the source
      array.
    width: The width of the sub-array to copy.
    height: The height of the sub-array to copy.
    dst_array: The destination ndarray to copy the sub-array into.
    dst_x: The X coordinate of the starting point to copy to, in the destination
      array.
    dst_y: The Y coordinate of the starting point to copy, in the destination
      array.

  Raises:
    SectionOutOfImageBoundsError if the sub-array to copy is out of scope of the
    source array or the destination array.
  """
  src_height, src_width = src_array.shape[:2]
  dst_height, dst_width = dst_array.shape[:2]
  if (
      src_x < 0
      or src_y < 0
      or src_x + width > src_width
      or src_y + height > src_height
  ):
    raise ez_wsi_errors.SectionOutOfImageBoundsError(
        'The sub-array to copy is out of the scope of the source array.'
    )
  if (
      dst_x < 0
      or dst_y < 0
      or dst_x + width > dst_width
      or dst_y + height > dst_height
  ):
    raise ez_wsi_errors.SectionOutOfImageBoundsError(
        'The sub-array to copy is out of the scope of the destination array.'
    )
  dst_array[dst_y : (dst_y + height), dst_x : (dst_x + width)] = src_array[
      src_y : (src_y + height), src_x : (src_x + width)
  ]


class DicomSlide:
  """Represents a DICOM pathology slide stored in a DICOMStore."""

  def __init__(
      self,
      dwi: dicom_web_interface.DicomWebInterface,
      path: dicom_path.Path,
      enable_client_slide_frame_decompression: bool,
      accession_number: Optional[str] = None,
      frame_cache_size: int = 128,
  ):
    """Constructor.

    Args:
      dwi: The DicomWebInterface that has been configured to be able to access
        the series referenced by the input path.
      path: The path to a DICOM series object in a DICOMStore.
      enable_client_slide_frame_decompression: Set to True to enable client side
        frame decompression optimization (Recommended value = True); remove
        parameter following GG validation.
      accession_number: The accession_number of the slide.
      frame_cache_size: Size of the LRU cache for retrieving frames.
    """
    self._level_map = slide_level_map.SlideLevelMap(dwi.get_instances(path))
    self._dwi = dwi
    self.path = path
    self.accession_number = accession_number
    native_level = _get_native_level(self._level_map)
    self.pixel_format = _get_pixel_format(native_level)
    self.native_magnification = _get_magnification(native_level)
    # Native height
    self.total_pixel_matrix_rows = native_level.height
    # Native width
    self.total_pixel_matrix_columns = native_level.width
    self._server_request_frame_lru_cache_size = frame_cache_size
    self._server_request_frame_cache = cachetools.LRUCache(frame_cache_size)
    self._enable_client_slide_frame_decompression = (
        enable_client_slide_frame_decompression
    )
    self._slide_frame_cache: Optional[
        abstract_slide_frame_cache.AbstractSlideFrameCache
    ] = None

  @property
  def slide_frame_cache(
      self,
  ) -> Optional[abstract_slide_frame_cache.AbstractSlideFrameCache]:
    return self._slide_frame_cache

  @slide_frame_cache.setter
  def slide_frame_cache(
      self,
      slide_frame_cache: abstract_slide_frame_cache.AbstractSlideFrameCache,
  ) -> None:
    self._slide_frame_cache = slide_frame_cache

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, DicomSlide):
      return str(self.path) == str(other.path)
    return False

  def __getstate__(self) -> MutableMapping[str, Any]:
    """Called to transform class state for pickeling.

    dicom_web_interface.DicomWebInterface is not pickled. When initalized the
    class is not compatiable with pickling. A inner member of the class
    DicomWebClientImpl holds credientials which are used to authenticate against
    the DICOM store. The interface needs to be re-initalized following pickling.

    The cache is not pickled. Two reasons, 1) It may be large and thereby
    consume significant time to pickle. 2) It is unclear if the the user/client
    who were to unpickle the data would have privlages to see the data that was
    pickled.

    Returns:
      Dict mapping attriubtes to values.
    """
    state = self.__dict__.copy()
    del state['_dwi']
    del state['_server_request_frame_cache']
    return state

  def __setstate__(self, dct: MutableMapping[str, Any]) -> None:
    """Called to init class from saved state."""
    self.__dict__ = dct
    self._dwi = None
    self._server_request_frame_cache = cachetools.LRUCache(
        self._server_request_frame_lru_cache_size
    )

  @property
  def dwi(self) -> dicom_web_interface.DicomWebInterface:
    return self._dwi

  @dwi.setter
  def dwi(self, val: dicom_web_interface.DicomWebInterface) -> None:
    self._dwi = val

  @property
  def magnifications(self) -> list[magnification_module.Magnification]:
    """Lists all Magnifications in the DicomSlide."""
    return [
        magnification_module.Magnification.FromPixelSize(
            level.pixel_spacing_x_mm * 1000
        )
        for level in self._level_map.level_map.values()
    ]

  def get_level_by_magnification(
      self, magnification: magnification_module.Magnification
  ) -> Optional[slide_level_map.Level]:
    """Gets the level corresponding to the input magnification.

    Args:
      magnification: The magnification to use for level lookup.

    Returns:
      The level corresponding to the input magnification. None if the requested
      magnification level does not exist.
    """
    # Converts pixel spacing returned by Magnification.NominalPixelSize() from
    # micrometers to millimeters.
    pixel_spacing_mm = magnification.nominal_pixel_size / 1000
    return self._level_map.get_level_by_pixel_spacing(pixel_spacing_mm)

  def _get_cached_frame_bytes(
      self,
      instance: slide_level_map.Instance,
      frame_number: int,
  ) -> Optional[bytes]:
    """Returns frame bytes from frame cache if possible.

    Args:
      instance: Instance to return frame from.
      frame_number: Instance frame number to return.

    Returns:
      Frame bytes or None.
    """
    if (
        self._slide_frame_cache is None
        or not is_client_side_pixel_decoding_supported(
            instance.dicom_object.get_value(tags.TRANSFER_SYNTAX_UID)
        )
    ):
      return None
    return self._slide_frame_cache.get_frame(
        str(instance.dicom_object.path), instance.frame_count, frame_number
    )

  def _get_frame_bytes_from_server(
      self,
      instance: slide_level_map.Instance,
      instance_frame_index: int,
      transcoding: dicom_web_interface.TranscodeDicomFrame,
  ) -> bytes:
    """Returns frame bytes from server.

    Args:
      instance: DICOM instance.
      instance_frame_index: Frame index to retrieve.
      transcoding: How to transcode DICOM frames.

    Returns:
      Frame bytes.
    """
    instance_path = instance.dicom_object.path
    cache_key = (
        f'i:{str(instance_path)} f:{instance_frame_index} t:{transcoding.value}'
    )
    frame_raw_bytes = self._server_request_frame_cache.get(cache_key)
    if frame_raw_bytes is not None:
      return frame_raw_bytes
    frame_raw_bytes = self.dwi.get_frame_image(
        instance_path, instance_frame_index, transcoding
    )
    self._server_request_frame_cache[cache_key] = frame_raw_bytes
    return frame_raw_bytes

  def _get_frame_client_transcoding(
      self,
      instance: slide_level_map.Instance,
      frame_number: int,
  ) -> Optional[np.ndarray]:
    """Returns DICOM Frame using DICOM server, transcodes to raw on server.

    Args:
      instance: DICOM instance within level to return frame from.
      frame_number: Frame number within the instance to return.

    Returns:
      Tuple[numpy array containing frame data, bool indicating if data should
      be cached True in LRU or is already cached within the level]
    """
    compressed_bytes = self._get_cached_frame_bytes(
        instance,
        frame_number,
    )
    if compressed_bytes is None:
      compressed_bytes = self._get_frame_bytes_from_server(
          instance,
          frame_number,
          dicom_web_interface.TranscodeDicomFrame.DO_NOT_TRANSCODE,
      )
    return dicom_frame_decoder.decode_dicom_compressed_frame_bytes(
        compressed_bytes
    )

  def _get_frame_server_transcoding(
      self,
      level: slide_level_map.Level,
      instance: slide_level_map.Instance,
      frame_number: int,
  ) -> np.ndarray:
    """Returns DICOM Frame using DICOM server, transcodes to raw on server.

    Args:
      level: WSI pyramid level.
      instance: DICOM instance within level to return frame from.
      frame_number: Frame number within the instance to return.

    Returns:
      Tuple[numpy array containing frame data, bool indicating if data should
      be cached True in LRU or is already cached within the level]
    """
    # Only use cache if frame bytes are stored a uncompressed little endian.
    if (
        level.transfer_syntax_uid
        not in _SUPPORTED_CLIENT_SIDE_DECODING_RAW_TRANSFER_SYNTAXS
    ):
      frame_raw_bytes = None
    else:
      frame_raw_bytes = self._get_cached_frame_bytes(
          instance,
          frame_number,
      )
    if frame_raw_bytes is None:
      frame_raw_bytes = self._get_frame_bytes_from_server(
          instance,
          frame_number,
          dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN,
      )
    return np.frombuffer(frame_raw_bytes, self.pixel_format).reshape(
        (level.frame_height, level.frame_width, level.samples_per_pixel)
    )

  def get_frame(
      self, magnification: magnification_module.Magnification, frame_number: int
  ) -> Optional[Frame]:
    """Gets a frame at a specific magnification level.

    The DICOMWeb API serves image pixels by the unit of frames. Frames have
    fixed size (width and height). Call get_patch() instead if you want to get
    an image patch at a specific loation, or with a specific dimension.

    The function utilizes a LRUCache to cache the most recent used frames.

    Args:
      magnification: The magnification level to fetch the frame from.
      frame_number: The frame number to be fetched. The frames are stored in
        arrays with 1-based indexing.

    Returns:
      Returns the requested frame if exists, None otherwise.

    Raises:
      InputFrameNumberOutOfRangeError if the input frame_number is
      out of range.
    """
    level = self.get_level_by_magnification(magnification)
    if level is None:
      return None
    if (
        frame_number < level.frame_number_min
        or frame_number > level.frame_number_max
    ):
      raise ez_wsi_errors.InputFrameNumberOutOfRangeError(
          f'frame_number value [{frame_number}] is out of range: '
          f'[{level.frame_number_min}, {level.frame_number_max}]'
      )
    instance = level.get_instance_by_frame(frame_number)
    if instance is None:
      # A frame may not exist in DICOMStore if it contains no tissue pixels.
      return None
    instance_frame_number = instance.frame_index_from_frame_number(frame_number)
    pos_x, pos_y = level.get_frame_position(frame_number)
    frame_ndarray = None
    if (
        self._enable_client_slide_frame_decompression
        and dicom_frame_decoder.can_decompress_dicom_transfer_syntax(
            level.transfer_syntax_uid
        )
    ):
      frame_ndarray = self._get_frame_client_transcoding(
          instance, instance_frame_number
      )
      # if frame_ndarray == None unable to decode bytes, likely cause is actual
      # pixel encoding does not match DICOM transfer syntax. Fail over to server
      # side transcoding.
    if frame_ndarray is None:
      frame_ndarray = self._get_frame_server_transcoding(
          level, instance, instance_frame_number
      )
    frame = Frame(
        x_origin=pos_x,
        y_origin=pos_y,
        width=level.frame_width,
        height=level.frame_height,
        image_np=frame_ndarray,
    )
    return frame

  def get_image(
      self, magnification: magnification_module.Magnification
  ) -> Image:
    """Gets an image from a specific magnification."""
    return Image(magnification, self)

  def get_patch(
      self,
      magnification: magnification_module.Magnification,
      x: int,
      y: int,
      width: int,
      height: int,
  ) -> Patch:
    """Gets a patch from a specific magnification level of the slide.

    The area of a patch is defined by its position(x, y) and its dimension
    (width, height). The position of a patch corresponds to the first pixel in
    the patch and has the smallest coordinates. All coordinates(x, y, width and
    height) are defined in the pixel space of the requested magnification level.

    This routine creates the requested patch on-the-fly, by sampling image
    pixels from all frames that are overlapped with the patch.

    Args:
      magnification: The magnification level to fetch the frame from.
      x: The X coordinate of the patch position.
      y: The Y coordinate of the patch position.
      width: The width of the patch.
      height: The height of the patch.

    Returns:
      Returns the requested patch if exists, None otherwise.

    Raises:
      MagnificationLevelNotFoundError if the requested level does not exist, or
      if the requested patch is out of the scope of the image.
    """
    level = self.get_level_by_magnification(magnification)
    if level is None:
      raise ez_wsi_errors.MagnificationLevelNotFoundError(
          f'The requested magnification level {magnification.as_string} '
          'does not exist.'
      )
    patch = Patch(magnification, x, y, width, height, slide=self)
    return patch

  def get_patch_bounds_dicom_instance_frame_numbers(
      self,
      mag: magnification_module.Magnification,
      patch_bounds_list: List[PatchBounds],
  ) -> Mapping[str, List[int]]:
    """Returns Map[DICOM instances: frame numbers] that fall in patch bounds.

    Args:
      mag: Magnification patch bounding list was generated at.
      patch_bounds_list: List of PatchBounds to return frame indexes for.

    Returns:
      Mapping between DICOM instances, path, and list of frames numbers required
      to render patches.
    """
    level = self.get_level_by_magnification(mag)
    if level is None:
      return {}
    slide_instance_frame_map = {}
    indexes_required_for_inference = []
    for patch_bounds in patch_bounds_list:
      patch = self.get_patch(
          mag,
          patch_bounds.x_origin,
          patch_bounds.y_origin,
          patch_bounds.width,
          patch_bounds.height,
      )
      # Frame indexes returns a list of indexes in the patch. Indexes
      # are returned in sorted order.
      indexes_required_for_inference.append(patch.frame_number())
    instance_frame_number_buffer = []
    instance = None
    # Use Heapq to merge pre-sorted lists into single sorted list
    # Result of heapq.merge can have duplicates
    for index in heapq.merge(*indexes_required_for_inference):
      if (
          instance is None
          or index - instance.frame_offset >= instance.frame_count
      ):
        if instance_frame_number_buffer:
          slide_instance_frame_map[str(instance.dicom_object.path)] = (
              instance_frame_number_buffer
          )
        instance = level.get_instance_by_frame(index)
        if instance is None:
          instance_frame_number_buffer = []
          continue
        instance_frame_number_buffer = [
            instance.frame_index_from_frame_number(index)
        ]
        continue

      instance_frame_number = instance.frame_index_from_frame_number(index)

      if (
          instance_frame_number_buffer
          and instance_frame_number_buffer[-1] == instance_frame_number
      ):
        # remove duplicates
        continue
      instance_frame_number_buffer.append(instance_frame_number)
    if instance_frame_number_buffer:
      slide_instance_frame_map[str(instance.dicom_object.path)] = (
          instance_frame_number_buffer
      )
    return slide_instance_frame_map

  @property
  def levels(self) -> Iterator[slide_level_map.Level]:
    """Returns iterator that contains all of a slide's DICOM Levels."""
    return iter(self._level_map.level_map.values())



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for pixel_spacing."""
from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import pixel_spacing


class PixelSpacingTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='1.1_percent_different',
          row_spacing=100,
          column_spacing=98.9,
      ),
      dict(
          testcase_name='1.1_percent_different_inverted',
          row_spacing=98.9,
          column_spacing=100,
      ),
      dict(
          testcase_name='10_percent_different',
          row_spacing=100,
          column_spacing=90,
      ),
      dict(
          testcase_name='10_percent_different_inverted',
          row_spacing=90,
          column_spacing=100,
      ),
  )
  def test_non_square_pixels(self, row_spacing: float, column_spacing: float):
    with self.assertRaises(ez_wsi_errors.NonSquarePixelError):
      pixel_spacing.PixelSpacing(row_spacing, column_spacing)

  @parameterized.named_parameters(
      dict(
          testcase_name='1_percent_different_pixels',
          row_spacing=100,
          column_spacing=99,
          expected_pixel_spacing=99.5,
          expected_row_spacing=100,
          expected_column_spacing=99,
      ),
      dict(
          testcase_name='1_percent_different_pixels_inverted',
          row_spacing=99,
          column_spacing=100,
          expected_pixel_spacing=99.5,
          expected_row_spacing=99,
          expected_column_spacing=100,
      ),
      dict(
          testcase_name='square_pixels',
          row_spacing=0.01,
          column_spacing=0.01,
          expected_pixel_spacing=0.01,
          expected_row_spacing=0.01,
          expected_column_spacing=0.01,
      ),
      dict(
          testcase_name='small_square_pixels',
          row_spacing=0.000001,
          column_spacing=0.0000010001,
          expected_pixel_spacing=0.00000100005,
          expected_row_spacing=0.000001,
          expected_column_spacing=0.0000010001,
      ),
  )
  def test_square_pixels(
      self,
      row_spacing: float,
      column_spacing: float,
      expected_pixel_spacing: float,
      expected_row_spacing: float,
      expected_column_spacing: float,
  ):
    ps = pixel_spacing.PixelSpacing(row_spacing, column_spacing)

    self.assertAlmostEqual(expected_pixel_spacing, ps.pixel_spacing_mm)
    self.assertAlmostEqual(expected_row_spacing, ps.row_spacing_mm)
    self.assertAlmostEqual(expected_column_spacing, ps.column_spacing_mm)

  @parameterized.named_parameters(
      dict(
          testcase_name='large_pixel_spacing',
          pixel_spacing_str='100.5',
          expected_pixel_spacing=100.5,
      ),
      dict(
          testcase_name='small_pixel_spacing',
          pixel_spacing_str='0.00000100005',
          expected_pixel_spacing=0.00000100005,
      ),
  )
  def test_from_string(
      self,
      pixel_spacing_str: str,
      expected_pixel_spacing: float,
  ):
    ps = pixel_spacing.PixelSpacing.FromString(pixel_spacing_str)

    self.assertEqual(expected_pixel_spacing, ps.pixel_spacing_mm)

  @parameterized.named_parameters(
      dict(
          testcase_name='5x_zoom',
          magnification_str='5x',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.002, 0.002),
          equal=True,
      ),
      dict(
          testcase_name='5X_zoom',
          magnification_str='5X',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.002, 0.002),
          equal=True,
      ),
      dict(
          testcase_name='5_zoom',
          magnification_str='5',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.002, 0.002),
          equal=True,
      ),
      dict(
          testcase_name='2.5X_zoom',
          magnification_str='2.5X',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.004, 0.004),
          equal=True,
      ),
      dict(
          testcase_name='2.5_zoom',
          magnification_str='2.5',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.004, 0.004),
          equal=True,
      ),
      dict(
          testcase_name='200_zoom',
          magnification_str='200',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.00005, 0.00005),
          equal=True,
      ),
      dict(
          testcase_name='5.1_zoom',
          magnification_str='5.1',
          expected_pixel_spacing=pixel_spacing.PixelSpacing(0.002, 0.002),
          equal=False,
      ),
  )
  def test_from_magnification_string(
      self,
      magnification_str: str,
      expected_pixel_spacing: pixel_spacing.PixelSpacing,
      equal: bool,
  ):
    ps = pixel_spacing.PixelSpacing.FromMagnificationString(magnification_str)

    self.assertEqual(equal, expected_pixel_spacing == ps)

  @parameterized.named_parameters(
      dict(
          testcase_name='all_dots',
          magnification_str='....',
      ),
      dict(
          testcase_name='one_dot',
          magnification_str='.',
      ),
      dict(
          testcase_name='weird_dots',
          magnification_str='.X5.',
      ),
      dict(
          testcase_name='wrong_order',
          magnification_str='X5',
      ),
      dict(
          testcase_name='negative',
          magnification_str='-5',
      ),
      dict(
          testcase_name='zero',
          magnification_str='0.0',
      ),
  )
  def test_from_magnification_string_error(
      self,
      magnification_str: str,
  ):
    with self.assertRaises(ez_wsi_errors.InvalidMagnificationStringError):
      pixel_spacing.PixelSpacing.FromMagnificationString(magnification_str)

  @parameterized.named_parameters(
      dict(
          testcase_name='large_pixel_spacing',
          float_pixel_spacing=100.5,
          expected_pixel_spacing=100.5,
      ),
      dict(
          testcase_name='small_pixel_spacing',
          float_pixel_spacing=0.00000100005,
          expected_pixel_spacing=0.00000100005,
      ),
  )
  def test_from_double(
      self,
      float_pixel_spacing: float,
      expected_pixel_spacing: float,
  ):
    ps = pixel_spacing.PixelSpacing.FromDouble(float_pixel_spacing)

    self.assertEqual(expected_pixel_spacing, ps.pixel_spacing_mm)

  @parameterized.named_parameters(
      dict(
          testcase_name='large_pixel_spacings_equal',
          first_pixel_spacing=100.5,
          second_pixel_spacing=100.5,
      ),
      dict(
          testcase_name='small_pixel_spacings_equal',
          first_pixel_spacing=0.00000100005,
          second_pixel_spacing=0.00000100005,
      ),
      dict(
          testcase_name='small_pixel_spacings_roughly_equal',
          first_pixel_spacing=0.00000100005,
          second_pixel_spacing=0.0000010000505,
      ),
  )
  def test_from_double_equal(
      self,
      first_pixel_spacing: float,
      second_pixel_spacing: float,
  ):
    ps_one = pixel_spacing.PixelSpacing.FromDouble(first_pixel_spacing)
    ps_two = pixel_spacing.PixelSpacing.FromDouble(second_pixel_spacing)

    self.assertEqual(ps_one, ps_two)

  @parameterized.named_parameters(
      dict(
          testcase_name='large_pixel_spacings_not_equal',
          first_pixel_spacing=100.5,
          second_pixel_spacing=102.0,
      ),
      dict(
          testcase_name='small_pixel_spacings_not_equal',
          first_pixel_spacing=0.000001100005,
          second_pixel_spacing=0.00000100005,
      ),
  )
  def test_from_double_not_equal(
      self, first_pixel_spacing: float, second_pixel_spacing: float
  ):
    ps_one = pixel_spacing.PixelSpacing.FromDouble(first_pixel_spacing)
    ps_two = pixel_spacing.PixelSpacing.FromDouble(second_pixel_spacing)

    self.assertNotEqual(ps_one, ps_two)

  @parameterized.named_parameters(
      dict(
          testcase_name='hashes_equal',
          first_row_spacing=100,
          first_column_spacing=99,
          second_row_spacing=100,
          second_column_spacing=99,
          expected_result=True,
      ),
      dict(
          testcase_name='hashes_not_equal',
          first_row_spacing=100,
          first_column_spacing=99,
          second_row_spacing=99,
          second_column_spacing=100,
          expected_result=False,
      ),
      dict(
          testcase_name='hashes_really_not_equal',
          first_row_spacing=100,
          first_column_spacing=99,
          second_row_spacing=99,
          second_column_spacing=99,
          expected_result=False,
      ),
  )
  def test_hashes_equal(
      self,
      first_row_spacing: float,
      first_column_spacing: float,
      second_row_spacing: float,
      second_column_spacing: float,
      expected_result: bool,
  ):
    ps_one = pixel_spacing.PixelSpacing(first_row_spacing, first_column_spacing)
    ps_two = pixel_spacing.PixelSpacing(
        second_row_spacing, second_column_spacing
    )

    self.assertEqual(ps_one.__hash__() == ps_two.__hash__(), expected_result)

  @parameterized.named_parameters(
      dict(
          testcase_name='string_not_equal',
          first_pixel_spacing=100.5,
          second_object='hello',
      ),
      dict(
          testcase_name='double_not_equal',
          first_pixel_spacing=100.5,
          second_object=100.5,
      ),
  )
  def test_from_double_equal_not_ps(
      self,
      first_pixel_spacing: float,
      second_object: object,
  ):
    ps_one = pixel_spacing.PixelSpacing.FromDouble(first_pixel_spacing)

    self.assertFalse(ps_one.__eq__(second_object))

  @parameterized.named_parameters(
      dict(
          testcase_name='large_zoom',
          float_pixel_spacing=0.0001,
          expected_magnification_string='100X',
      ),
      dict(
          testcase_name='80x_zoom',
          float_pixel_spacing=0.000125,
          expected_magnification_string='80X',
      ),
      dict(
          testcase_name='40x_zoom',
          float_pixel_spacing=0.00025,
          expected_magnification_string='40X',
      ),
      dict(
          testcase_name='20x_zoom',
          float_pixel_spacing=0.0005,
          expected_magnification_string='20X',
      ),
      dict(
          testcase_name='10x_zoom',
          float_pixel_spacing=0.001,
          expected_magnification_string='10X',
      ),
      dict(
          testcase_name='5x_zoom',
          float_pixel_spacing=0.002,
          expected_magnification_string='5X',
      ),
      dict(
          testcase_name='2.5x_zoom',
          float_pixel_spacing=0.004,
          expected_magnification_string='2.5X',
      ),
      dict(
          testcase_name='1.25x_zoom',
          float_pixel_spacing=0.008,
          expected_magnification_string='1.25X',
      ),
      dict(
          testcase_name='.625x_zoom',
          float_pixel_spacing=0.016,
          expected_magnification_string='0.625X',
      ),
      dict(
          testcase_name='.3125x_zoom',
          float_pixel_spacing=0.032,
          expected_magnification_string='0.3125X',
      ),
      dict(
          testcase_name='.15625x_zoom',
          float_pixel_spacing=0.064,
          expected_magnification_string='0.15625X',
      ),
      dict(
          testcase_name='.078125x_zoom',
          float_pixel_spacing=0.128,
          expected_magnification_string='0.078125X',
      ),
      dict(
          testcase_name='smallest_zoom',
          float_pixel_spacing=0.256,
          expected_magnification_string='0.0390625X',
      ),
  )
  def test_as_magnification_string(
      self,
      float_pixel_spacing: float,
      expected_magnification_string: str,
  ):
    ps = pixel_spacing.PixelSpacing.FromDouble(float_pixel_spacing)

    self.assertEqual(expected_magnification_string, ps.as_magnification_string)

  def test_pixel_spacing_consistency_from_double(self):
    # Test string -> PS is equal to string -> PS -> double -> PS
    self.assertEqual(
        pixel_spacing.PixelSpacing.FromMagnificationString('10X'),
        pixel_spacing.PixelSpacing.FromDouble(
            pixel_spacing.PixelSpacing.FromMagnificationString(
                '10X'
            ).pixel_spacing_mm
        ),
    )

  def test_pixel_spacing_consistency_from_string(self):
    # Test string -> PS is equal to string -> PS -> string -> PS
    self.assertEqual(
        pixel_spacing.PixelSpacing.FromMagnificationString('10X'),
        pixel_spacing.PixelSpacing.FromMagnificationString(
            pixel_spacing.PixelSpacing.FromMagnificationString(
                '10X'
            ).as_magnification_string
        ),
    )

  def test_scaling_factor(self):
    self.assertEqual(
        pixel_spacing.PixelSpacing(0.01, 0.01).as_magnification_string,
        pixel_spacing.PixelSpacing(0.02, 0.02, 0.02).as_magnification_string,
    )

  def test_scaling_factor_from_string(self):
    self.assertEqual(
        pixel_spacing.PixelSpacing.FromString('.01').as_magnification_string,
        pixel_spacing.PixelSpacing.FromString(
            '.02', 0.02
        ).as_magnification_string,
    )

  def test_string_scaling_factor_from_magnification(self):
    self.assertEqual(
        pixel_spacing.PixelSpacing.FromMagnificationString(
            '5X'
        ).as_magnification_string,
        pixel_spacing.PixelSpacing.FromMagnificationString(
            '5X', 0.02
        ).as_magnification_string,
    )

  def test_scaling_factor_from_double(self):
    self.assertEqual(
        pixel_spacing.PixelSpacing.FromDouble(0.01).as_magnification_string,
        pixel_spacing.PixelSpacing.FromDouble(
            0.02, 0.02
        ).as_magnification_string,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='1_percent_different_pixels',
          row_spacing=100,
          column_spacing=99,
          diff_tolerance=0.00001,
      ),
      dict(
          testcase_name='10_percent_different_pixels',
          row_spacing=100,
          column_spacing=90,
          diff_tolerance=0.09,
      ),
  )
  def test_non_square_pixels_with_diff_tolerance(
      self, row_spacing: float, column_spacing: float, diff_tolerance: float
  ):
    with self.assertRaises(ez_wsi_errors.NonSquarePixelError):
      pixel_spacing.PixelSpacing(
          row_spacing, column_spacing, spacing_diff_tolerance=diff_tolerance
      )

  @parameterized.named_parameters(
      dict(
          testcase_name='10_percent_different_pixels',
          row_spacing=100,
          column_spacing=90,
          diff_tolerance=0.1,
          expected_pixel_spacing=95,
          expected_row_spacing=100,
          expected_column_spacing=90,
      ),
      dict(
          testcase_name='10_percent_different_pixels_inverted',
          row_spacing=90,
          column_spacing=100,
          diff_tolerance=0.1,
          expected_pixel_spacing=95,
          expected_row_spacing=90,
          expected_column_spacing=100,
      ),
  )
  def test_square_pixels_with_diff_tolerance(
      self,
      row_spacing: float,
      column_spacing: float,
      diff_tolerance: float,
      expected_pixel_spacing: float,
      expected_row_spacing: float,
      expected_column_spacing: float,
  ):
    ps = pixel_spacing.PixelSpacing(
        row_spacing, column_spacing, spacing_diff_tolerance=diff_tolerance
    )

    self.assertAlmostEqual(expected_pixel_spacing, ps.pixel_spacing_mm)
    self.assertAlmostEqual(expected_row_spacing, ps.row_spacing_mm)
    self.assertAlmostEqual(expected_column_spacing, ps.column_spacing_mm)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Error classes for EZ WSI DicomWeb."""


class EZWsiError(Exception):
  pass


class CoordinateOutofImageDimensionsError(EZWsiError):
  pass


class DicomImageMissingRegionError(EZWsiError):
  pass


class DicomPathError(EZWsiError):
  pass


class DicomSlideMissingError(EZWsiError):
  pass


class DicomTagNotFoundError(EZWsiError):
  pass


class FrameNumberOutofBoundsError(EZWsiError):
  pass


class InputFrameNumberOutOfRangeError(EZWsiError):
  pass


class InvalidDicomTagError(EZWsiError):
  pass


class InvalidMagnificationStringError(EZWsiError):
  pass


class MagnificationLevelNotFoundError(EZWsiError):
  pass


class NoDicomLevelsDetectedError(EZWsiError):
  pass


class NonSquarePixelError(EZWsiError):
  pass


class PatchIntersectionNotFoundError(EZWsiError):
  pass


class SectionOutOfImageBoundsError(EZWsiError):
  pass


class SlideLevelContainsInstancesWithDifferentTransferSyntaxUIDError(
    EZWsiError
):

  def __init__(self):
    super().__init__(
        'DICOM level instances do not have the same TransferSyntaxUID'
    )


class UnexpectedDicomObjectInstanceError(EZWsiError):
  pass


class UnexpectedDicomSlideCountError(EZWsiError):
  pass


class UnsupportedPixelFormatError(EZWsiError):
  pass



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Dicom Web abstraction layer."""
from ez_wsi_dicomweb import dicom_web_interface
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb.dicom_slide import DicomSlide

from hcls_imaging_ml_toolkit import dicom_path
from hcls_imaging_ml_toolkit import dicom_web


class DicomStore:
  """Abstraction on top of DicomWeb.

  Designed to be the layer of internal PathDB. Provides the single point of
  access to the hierarchy of slide/image/patch.
  """

  def __init__(
      self,
      dicomstore_path: str,
      enable_client_slide_frame_decompression: bool,
      dicomweb=None,
  ):
    self._enable_client_slide_frame_decompression = (
        enable_client_slide_frame_decompression
    )
    self.dicomstore_path = dicomstore_path
    self.dicomweb = dicomweb
    if dicomweb is None:
      dicom_web_interface.DicomWebInterface(dicom_web.DicomWebClientImpl())

  def get_slide_by_accession_number(self, accession_number: str) -> DicomSlide:
    """Searches a DicomSlide object by accession number.

    Args:
      accession_number: DICOM tag, e.g. AccessionNumber=123.

    Returns:
      A DicomSlide object.

    Raises:
      UnexpectedDicomSlideCountError if the slide count for the series is not 1.
    """
    dicom = self.dicomweb.get_series(
        dicom_path.FromString(self.dicomstore_path),
        {'AccessionNumber': accession_number},
    )
    if len(dicom) != 1:
      raise ez_wsi_errors.UnexpectedDicomSlideCountError(
          f'Expect single slide for {accession_number}, len(dicom)={len(dicom)}'
      )

    return DicomSlide(
        self.dicomweb,
        dicom[0].path,
        enable_client_slide_frame_decompression=self._enable_client_slide_frame_decompression,
        accession_number=accession_number,
    )

  def get_slide(self, study_id_series_id: str) -> DicomSlide:
    """Gets a DicomSlide object.

    Args:
      study_id_series_id: Colon ':' concatenated study Id and series Id.

    Returns:
      A DicomSlide object.

    Raises:
      DicomSlideMissingError if the slide is not constructed correctly.
    """
    store_path = dicom_path.FromString(self.dicomstore_path)
    study_uid, series_uid = study_id_series_id.split(':')
    series_path = dicom_path.FromPath(
        store_path, study_uid=study_uid, series_uid=series_uid
    )
    slide = DicomSlide(
        self.dicomweb,
        series_path,
        enable_client_slide_frame_decompression=self._enable_client_slide_frame_decompression,
    )

    if slide is None:
      raise ez_wsi_errors.DicomSlideMissingError(
          f'Error constructing DicomSlide for slide {study_id_series_id}.'
      )
    return slide



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""SlideLevelMap defines the relationship between frames and DICOM instances."""
import collections
import dataclasses
import math
from typing import Collection, Dict, Iterator, Mapping, Optional, Tuple

from ez_wsi_dicomweb import dicom_web_interface
from ez_wsi_dicomweb import ez_wsi_errors

from hcls_imaging_ml_toolkit import dicom_path
from hcls_imaging_ml_toolkit import tags

# Since the level distance is a log function, this max distance value allows
# a level that has a pixel spacing ratio within range of [2^-0.3, 2^0.3] to be
# matched as the closest level by get_level_by_pixel_spacing() function.
_MAX_LEVEL_DIST = 0.3


@dataclasses.dataclass(frozen=True)
class Instance:
  """Wrapper of a DICOM instance object, with details for frame lookup.

  Used by the Level class to represent an instance in a specific
  magnification level.

  Attributes:
    frame_offset: The frame index offset of the first possible frame from the
      instance on the parent magnification level.
    frame_count: Number of frames contained inside the instance.
    dicom_object: The DicomObject of the instance.
  """

  frame_offset: int
  frame_count: int
  dicom_object: dicom_web_interface.DicomObject

  def frame_index_from_frame_number(self, frame_number: int) -> int:
    """Converts a frame_number to a frame_index within this instance.

    The frame_number and frame_index are two different concepts:
      - The frame_number is global to a Level. The values of frame_number are
        within [level.frame_number_min and level.frame_number_max].
      - The frame_index is local to the host instance. The frame_index starts
        with 1. The following equation converts a frame_number to frame_index:
        frame_index = frame_number - instance.frame_offset + 1

    Args:
     frame_number: The frame_number to be converted from.

    Returns:
     The index of the frame within this instance.
    """
    return frame_number - self.frame_offset + 1


@dataclasses.dataclass(frozen=True)
class Level:
  """Represents the dimensions and instances of a specific magnification level.

  Attributes:
    level_index: The index of the level using 1-based indexing. A level with a
      higher index corresponding to a lower magnification level.
    width: The width, in number of pixels, of the entire level.
    height: The height, in number of pixels, of the entire level.
    samples_per_pixel: The number of samples per pixel. i.e. 3 for RGB format.
    bits_allocated: The number of bits allocated for storing one pixel.
    high_bit: The index of the highest bit in each sample of a pixel.
    pixel_spacing_x_mm: The pixel spacing, in millimeters, in x direction of the
      level.
    pixel_spacing_y_mm: The pixel spacing, in millimeters, in y direction of the
      level.
    frame_width: The width, in pixels, of all frames on this level.
    frame_height: The width, in pixels, of all frames on this level.
    frame_number_min: The minimum frame number at this level.
    frame_number_max: The maximum frame number at this level.
    instances: All instances on this level, keyed by the frame_offset of the
      instance. The instances should be ordered in the dictionary by the key.
    transfer_syntax_uid: DICOM transfer syntax instances are encoded as.
  """

  level_index: int
  width: int
  height: int
  samples_per_pixel: int
  bits_allocated: int
  high_bit: int
  pixel_spacing_x_mm: float
  pixel_spacing_y_mm: float
  frame_width: int
  frame_height: int
  frame_number_min: int
  frame_number_max: int
  instances: Dict[int, Instance]
  transfer_syntax_uid: str

  def get_instance_by_frame(self, frame_number: int) -> Optional[Instance]:
    """Gets the instance that contains the requested frame.

    Args:
      frame_number: The frame number of the frame to request. The frame number
        of a frame is stored in the DICOM tag
        CONCATENATION_FRAME_OFFSET_NUMBER('00209228').

    Returns:
      The instance that contains the requested frame, or None if the input frame
      is out of range of any instance.
    """
    # TODO: replace with binary search if performance is an issue.
    for frame_offset, instance in self.instances.items():
      if frame_number < frame_offset:
        break
      if frame_number < frame_offset + instance.frame_count:
        return instance
    return None

  def get_frame_number_by_point(self, x: int, y: int) -> int:
    """Gets the frame number corresponding to the input coordinate.

    Args:
      x: The x coordinate at the level the instance belongs to.
      y: The y coordinate at the level the instance belongs to.

    Returns:
      The frame number that contains the input coordinate.

    Raises:
      CoordinateOutofImageDimensionsError if the input coordinate is out of the
      range.
    """
    if x < 0 or y < 0 or x >= self.width or y >= self.height:
      raise ez_wsi_errors.CoordinateOutofImageDimensionsError(
          f'The input coordinate {x, y} is out of the range: '
          f'{0, 0, self.width - 1, self.height - 1}'
      )
    frame_x = int(x / self.frame_width)
    frame_y = int(y / self.frame_height)
    frames_per_row = int(math.ceil(float(self.width) / float(self.frame_width)))
    return frame_y * frames_per_row + frame_x

  def get_frame_position(self, frame_number: int) -> Tuple[int, int]:
    """Gets the coordinate of the upper-left corner of the input frame.

    Args:
      frame_number: The input frame number to get the position of.

    Returns:
      The X and Y coordinates of the upper-left corner of the input frame.

    Raises:
      FrameNumberOutofBounds if the input frame number is out of the
      range.
    """
    if (
        frame_number < self.frame_number_min
        or frame_number > self.frame_number_max
    ):
      raise ez_wsi_errors.FrameNumberOutofBoundsError(
          f'The input frame number ({frame_number}) is out of the range: '
          f'{self.frame_number_min, self.frame_number_max}'
      )
    frame_number = int(frame_number)
    frames_per_row = int(math.ceil(float(self.width) / float(self.frame_width)))
    frame_x = frame_number % frames_per_row
    frame_y = int(frame_number / frames_per_row)
    return frame_x * self.frame_width, frame_y * self.frame_height

  def get_instance_by_point(self, x: int, y: int) -> Optional[Instance]:
    """Gets the instance that contains the input point.

    Args:
      x: The x coordinate at the level the instance belongs to.
      y: The y coordinate at the level the instance belongs to.

    Returns:
      The instance that contains the input point, or None if the point is out of
      range of the level or is not within any instance.
    """
    return self.get_instance_by_frame(self.get_frame_number_by_point(x, y))

  @property
  def instance_iterator(self) -> Iterator[Instance]:
    """Returns iterator of level's DICOM instances."""
    return iter(self.instances.values())


def _build_level_map(
    dicom_objects: Collection[dicom_web_interface.DicomObject],
) -> Mapping[int, Level]:
  """Returns a slide level map built from a set of input DICOM instances.

  The algorithm groups instances in multiple sets by level. The result is stored
  in a ordered dictionary, using the level index as the key

  Within each level, all instances belonging to that level are stored into a
  dictionary, using the frame offset of the instance as the key:
    Dict[frame_offset, Instance]

  Both the level dictionary and the instance dictionary need to be accessed in
  order by the key values. To avoid sorting the keys on-the-fly, we sort
  both dictionaries beforehand.

  Args:
    dicom_objects: DICOM objects to use to build the SlideLevelMap.

  Returns:
    A sorted dict that contains all levels of the DICOM object.

  Raises:
    UnexpectedDicomObjectInstanceError: If any of the input objects is not a
    DICOM instance object.
    NoDicomLevelsDetectedError: No level was detected in the Dicom object.
    DicomTagNotFoundError: If any of the required tags is missing from any DICOM
    object.
    InvalidDicomTagError: If check_no_zero is enabled and the tag has zero
    value.
    SlideLevelContainsInstancesWithDifferentTransferSyntaxUIDError: Slide level
      contains instances with different transfer syntaxes.
  """
  levels = {}
  # https://dicom.nema.org/dicom/2013/output/chtml/part03/sect_A.32.html#sect_A.32.8
  vl_micro_sop_class_id = "1.2.840.10008.5.1.4.1.1.77.1.6"
  level_instances = collections.defaultdict(dict)
  for dicom_object in dicom_objects:
    if dicom_object.type() != dicom_path.Type.INSTANCE:
      raise ez_wsi_errors.UnexpectedDicomObjectInstanceError(
          'SlideLevelMap expects all input DicomObject to have a type of '
          f'INSTANCE. Actual: {dicom_object.type()}'
      )
    _check_for_tag_or_raise(dicom_object, tags.SOP_CLASS_UID)
    if dicom_object.get_value(tags.SOP_CLASS_UID) != vl_micro_sop_class_id:
      continue

    # Some clincal viewer (sectra) appears to require label, thumbnail,
    # and macro images (overviews) to be encoded using the
    # VL Whole Slide Microscopy Image Storage IOD to be visible in the software.
    image_type = dicom_object.get_list_value(tags.IMAGE_TYPE)
    if image_type:
      if frozenset(image_type) & frozenset(('LABEL', 'THUMBNAIL', 'OVERVIEW')):
        continue

    # Makes sure the following tags have defined values in dicom_object.
    _check_for_tag_or_raise(dicom_object, tags.INSTANCE_NUMBER)
    _check_for_tag_or_raise(dicom_object, tags.TOTAL_PIXEL_MATRIX_COLUMNS, True)
    _check_for_tag_or_raise(dicom_object, tags.TOTAL_PIXEL_MATRIX_ROWS, True)
    _check_for_tag_or_raise(dicom_object, tags.COLUMNS, True)
    _check_for_tag_or_raise(dicom_object, tags.ROWS, True)
    _check_for_tag_or_raise(dicom_object, tags.SAMPLES_PER_PIXEL, True)
    _check_for_tag_or_raise(dicom_object, tags.BITS_ALLOCATED, True)
    _check_for_tag_or_raise(dicom_object, tags.HIGH_BIT, True)
    _check_for_tag_or_raise(dicom_object, tags.IMAGE_VOLUME_WIDTH, True)
    _check_for_tag_or_raise(dicom_object, tags.IMAGE_VOLUME_HEIGHT, True)
    _check_for_tag_or_raise(dicom_object, tags.TRANSFER_SYNTAX_UID, True)

    level_index = dicom_object.get_value(tags.INSTANCE_NUMBER)

    level_index_data = levels.get(level_index)
    if level_index_data is not None:
      if level_index_data.transfer_syntax_uid != dicom_object.get_value(
          tags.TRANSFER_SYNTAX_UID
      ):
        raise ez_wsi_errors.SlideLevelContainsInstancesWithDifferentTransferSyntaxUIDError()
    else:
      level_index_data = Level(
          level_index=level_index,
          width=dicom_object.get_value(tags.TOTAL_PIXEL_MATRIX_COLUMNS),
          height=dicom_object.get_value(tags.TOTAL_PIXEL_MATRIX_ROWS),
          samples_per_pixel=dicom_object.get_value(tags.SAMPLES_PER_PIXEL),
          bits_allocated=dicom_object.get_value(tags.BITS_ALLOCATED),
          high_bit=dicom_object.get_value(tags.HIGH_BIT),
          pixel_spacing_x_mm=dicom_object.get_value(tags.IMAGE_VOLUME_WIDTH)
          / dicom_object.get_value(tags.TOTAL_PIXEL_MATRIX_COLUMNS),
          pixel_spacing_y_mm=dicom_object.get_value(tags.IMAGE_VOLUME_HEIGHT)
          / dicom_object.get_value(tags.TOTAL_PIXEL_MATRIX_ROWS),
          frame_width=dicom_object.get_value(tags.COLUMNS),
          frame_height=dicom_object.get_value(tags.ROWS),
          frame_number_min=-1,
          frame_number_max=-1,
          instances={},
          transfer_syntax_uid=dicom_object.get_value(tags.TRANSFER_SYNTAX_UID),
      )
      levels[level_index] = level_index_data

    frame_offset = (
        dicom_object.get_value(tags.CONCATENATION_FRAME_OFFSET_NUMBER) or 0
    )
    level_instances[level_index][frame_offset] = Instance(
        frame_offset=frame_offset,
        frame_count=dicom_object.get_value(tags.NUMBER_OF_FRAMES),
        dicom_object=dicom_object,
    )
  if not levels:
    raise ez_wsi_errors.NoDicomLevelsDetectedError(
        f'No level detected from the input DICOM objects {dicom_objects}'
    )

  # Sort levels based on level index.
  sorted_levels = collections.OrderedDict()
  for level_index in sorted(levels):
    # Sort instances in each level based on frame_offset
    sorted_level_index_data = levels[level_index]
    instances = level_instances[level_index]
    sorted_instances = collections.OrderedDict()
    for frame_offset in sorted(instances):
      sorted_instances[frame_offset] = instances[frame_offset]
    frame_number_min = next(iter(sorted_instances.keys()))
    max_frame_offset = next(reversed(sorted_instances.keys()))
    frame_number_max = (
        max_frame_offset + instances[max_frame_offset].frame_count
    )
    sorted_levels[level_index] = dataclasses.replace(
        sorted_level_index_data,
        frame_number_max=frame_number_max,
        frame_number_min=frame_number_min,
        instances=sorted_instances,
    )

  return sorted_levels


def _check_for_tag_or_raise(
    dicom_object: dicom_web_interface.DicomObject,
    tag: tags.DicomTag,
    check_no_zero: bool = False,
):
  """Check for the existence of a DICOM tag in a DICOM object.

  Args:
    dicom_object: The host DICOM object to check against.
    tag: The DICOM tag to check for.
    check_no_zero: If enabled (true), this method will check if the value of the
      tag is non-zero.

  Raises:
    DicomTagNotFoundError: If the tag does not exist in the object.
    InvalidDicomTagError: If check_no_zero is enabled and the tag has zero
    value.
  """
  value = dicom_object.get_value(tag)
  if value is None:
    raise ez_wsi_errors.DicomTagNotFoundError(
        f'DICOM tag {tag.number} is missing from the DICOM object: '
        f'{str(dicom_object.path)}'
    )
  if check_no_zero and value == 0:
    raise ez_wsi_errors.InvalidDicomTagError(
        f'DICOM tag {tag.number} cannot have zero value.'
    )


class SlideLevelMap(object):
  """A class that builds and maintains the level-instance-frame mapping.

  When converting a WSI slide to DICOM format, a single WSI slide gets mapped to
  one DICOM series, which stores images at multiple levels of magnification with
  multiple DICOM instances.
  At each level of magnification, the image is divided into multiple frames with
  a fixed frame size(i.e. 500).
  DICOM tries to store all frames at a magnification level into one DICOM
  instance. If the number of frames at that level exceeds the limit, which is
  typically 2048, multiple instances are used. In that case, the index of the
  first possible frame within an instance is tagged as the frame_offset of that
  instance.
  """

  def __init__(
      self, dicom_objects: Collection[dicom_web_interface.DicomObject]
  ):
    """Constructor.

    Args:
      dicom_objects: DICOM objects of all instances within a DICOM series.

    Raises:
      UnexpectedDicomObjectInstanceError: If any of the input objects is not a
      DICOM instance object
      NoDicomLevelsDetectedError: No level was detected in the Dicom object.
      DicomTagNotFoundError: If any of the required tags is missing from any
      DICOM object.
      InvalidDicomTagError: If check_no_zero is enabled and the tag has zero
      value.
      SlideLevelContainsInstancesWithDifferentTransferSyntaxUIDError: Slide
      level contains instances with different transfer syntaxes.
    """
    self._level_map = _build_level_map(dicom_objects)
    level_index_list = list(self._level_map)
    self.level_index_min = level_index_list[0]
    self.level_index_max = level_index_list[-1]

  @property
  def level_map(self) -> Mapping[int, Level]:
    """Returns an ordered level map built from a set of input DICOM instances.

    The algorithm groups instances in multiple sets by level. The result is
    stored in a dictionary, using the level index as the key:
      Dict[level_index, Level]
    """
    return self._level_map

  def get_level(self, level_index: int) -> Optional[Level]:
    """Returns the level by requested level index.

    Args:
      level_index: The level index to be required. Level index usually starts at
        1.
    """
    return self._level_map.get(level_index)

  def get_level_by_pixel_spacing(
      self, pixel_spacing_mm: float
  ) -> Optional[Level]:
    """Returns the level that has the closest pixel spacing as the input.

    NOTE: This code assumes all images will have square pixels. It also assumes
    magnifications are powers of 2.
    Args:
      pixel_spacing_mm: the requested pixel spacing in millimeters.

    Returns:
      The level that closely matches with the requestd pixel spacing, or None
      if the requested pixel spacing is out range of any existing levels.
    """

    def _compute_distance_from_level(index: int) -> float:
      """Returns the distance between requested pixel spacing and a level.

      The distance is defined as the log2 of the pixel spacing ratio, which is
      defined as:
        pixel_spacing_ratio = requested_pixel_spacing/level_pixel_spacing

      Args:
        index: The index of the level to be calculated against.
      """
      return abs(
          math.log2(
              pixel_spacing_mm / self._level_map[index].pixel_spacing_x_mm
          )
      )

    # TODO: replace with binary search if performance is an issue.
    min_level = None
    min_distance = 1e10
    # Finds the min_distance in level_map.
    for level in self._level_map:
      distance = _compute_distance_from_level(level)
      if distance < min_distance:
        min_distance = distance
        min_level = level

    if min_distance < _MAX_LEVEL_DIST:
      return self._level_map[min_level]

    return None



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.dicom_web_interface."""
from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import dicom_test_utils
from ez_wsi_dicomweb import dicom_web_interface
from ez_wsi_dicomweb import ez_wsi_errors
import mock

from hcls_imaging_ml_toolkit import dicom_path
from hcls_imaging_ml_toolkit import dicom_web
from hcls_imaging_ml_toolkit import tags


class DicomWebInterfaceTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._dicom_store_path = dicom_path.FromString(
        dicom_test_utils.TEST_STORE_PATH
    )
    # Mock the first set of DICOM paths:
    # studies/{_TEST_STUDY_1}
    self._study_path_1 = dicom_path.FromPath(
        self._dicom_store_path, study_uid=dicom_test_utils.TEST_STUDY_UID_1
    )
    # studies/{_TEST_STUDY_1}/series/{_TEST_SERIES_1}
    self._series_path_1 = dicom_path.FromPath(
        self._study_path_1, series_uid=dicom_test_utils.TEST_SERIES_UID_1
    )
    # studies/{_TEST_STUDY_1}/series/{_TEST_SERIES_1}/instances/{_TEST_INSTANCE_1}
    self._instance_path_1 = dicom_path.FromPath(
        self._series_path_1, instance_uid=dicom_test_utils.TEST_INSTANCE_UID_1
    )

    # Mock the second set of dicom paths:
    # studies/{_TEST_STUDY_2}
    self._study_path_2 = dicom_path.FromPath(
        self._dicom_store_path, study_uid=dicom_test_utils.TEST_STUDY_UID_2
    )
    # studies/{_TEST_STUDY_2}/series/{_TEST_SERIES_2}
    self._series_path_2 = dicom_path.FromPath(
        self._study_path_2, series_uid=dicom_test_utils.TEST_SERIES_UID_2
    )
    # studies/{_TEST_STUDY_2}/series/{_TEST_SERIES_2}/instances/{_TEST_INSTANCE_2}
    self._instance_path_2 = dicom_path.FromPath(
        self._series_path_2, instance_uid=dicom_test_utils.TEST_INSTANCE_UID_2
    )

  def test_dicom_object_constructor_with_invalid_input_raise_error(self):
    with self.assertRaises(ez_wsi_errors.DicomPathError):
      dicom_web_interface.DicomObject(self._dicom_store_path, {})

  def test_get_studies_with_invalid_input_raise_error(self):
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    with self.subTest(name='Study'):
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_studies(self._study_path_1)
    with self.subTest(name='Series'):
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_studies(self._series_path_1)
    with self.subTest(name='Instance'):
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_studies(self._instance_path_1)

  def test_get_studies_with_valid_input(self):
    all_studies_json = [
        {
            '0020000D': {
                'Value': [dicom_test_utils.TEST_STUDY_UID_1],
                'vr': 'UI',
            }
        },
        {
            '0020000D': {
                'Value': [dicom_test_utils.TEST_STUDY_UID_2],
                'vr': 'UI',
            }
        },
    ]
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.QidoRs.return_value = all_studies_json
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    studies = dwi.get_studies(self._dicom_store_path)
    self.assertCountEqual(
        studies,
        [
            dicom_web_interface.DicomObject(
                self._study_path_1, all_studies_json[0]
            ),
            dicom_web_interface.DicomObject(
                self._study_path_2, all_studies_json[1]
            ),
        ],
        (
            'All dicom objects returned should have a type of STUDY, with '
            'corresponding dicom path and original dicom tags attached.'
        ),
    )

  def test_get_series_with_invalid_input_raise_error(self):
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    invalid_path = [
        self._series_path_1,
        self._instance_path_1,
    ]
    for path in invalid_path:
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_series(path)

  def test_get_series_with_store_level_path(self):
    all_series_json = [
        {
            '0020000D': {
                'Value': [dicom_test_utils.TEST_STUDY_UID_1],
                'vr': 'UI',
            },
            '0020000E': {
                'Value': [dicom_test_utils.TEST_SERIES_UID_1],
                'vr': 'UI',
            },
        },
        {
            '0020000D': {
                'Value': [dicom_test_utils.TEST_STUDY_UID_2],
                'vr': 'UI',
            },
            '0020000E': {
                'Value': [dicom_test_utils.TEST_SERIES_UID_2],
                'vr': 'UI',
            },
        },
    ]
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.QidoRs.return_value = all_series_json
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    series = dwi.get_series(self._dicom_store_path)
    self.assertCountEqual(
        series,
        [
            dicom_web_interface.DicomObject(
                self._series_path_1, all_series_json[0]
            ),
            dicom_web_interface.DicomObject(
                self._series_path_2, all_series_json[1]
            ),
        ],
        (
            'All DICOM objects returned should have a type of SERIES, with '
            'corresponding DICOM path and original DICOM tags attached.'
        ),
    )

  def test_get_series_with_study_level_path(self):
    all_series_json = [
        {
            '0020000E': {
                'Value': [dicom_test_utils.TEST_SERIES_UID_1],
                'vr': 'UI',
            }
        }
    ]
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.QidoRs.return_value = all_series_json
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    instances = dwi.get_series(self._study_path_1)
    self.assertEqual(
        instances[0],
        dicom_web_interface.DicomObject(
            self._series_path_1, all_series_json[0]
        ),
        (
            'The returned DICOM object should have a type of SERIES, with '
            'corresponding DICOM path and original DICOM tags attached.'
        ),
    )

  def test_get_instances_with_invalid_input_raise_error(self):
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    with self.assertRaises(ez_wsi_errors.DicomPathError):
      dwi.get_instances(self._instance_path_1)

  def test_get_instances_with_store_level_path(self):
    all_instances_json = [
        {
            '0020000D': {
                'Value': [dicom_test_utils.TEST_STUDY_UID_1],
                'vr': 'UI',
            },
            '0020000E': {
                'Value': [dicom_test_utils.TEST_SERIES_UID_1],
                'vr': 'UI',
            },
            '00080018': {
                'Value': [dicom_test_utils.TEST_INSTANCE_UID_1],
                'vr': 'UI',
            },
        },
        {
            '0020000D': {
                'Value': [dicom_test_utils.TEST_STUDY_UID_2],
                'vr': 'UI',
            },
            '0020000E': {
                'Value': [dicom_test_utils.TEST_SERIES_UID_2],
                'vr': 'UI',
            },
            '00080018': {
                'Value': [dicom_test_utils.TEST_INSTANCE_UID_2],
                'vr': 'UI',
            },
        },
    ]
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.QidoRs.return_value = all_instances_json
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    instances = dwi.get_instances(self._dicom_store_path)
    self.assertCountEqual(
        instances,
        [
            dicom_web_interface.DicomObject(
                self._instance_path_1, all_instances_json[0]
            ),
            dicom_web_interface.DicomObject(
                self._instance_path_2, all_instances_json[1]
            ),
        ],
        (
            'All DICOM objects returned should have a type of INSTANCE, with '
            'corresponding DICOM path and original DICOM tags attached.'
        ),
    )

  def test_get_instances_with_study_level_path(self):
    all_instances_json = [{
        '0020000E': {'Value': [dicom_test_utils.TEST_SERIES_UID_1], 'vr': 'UI'},
        '00080018': {
            'Value': [dicom_test_utils.TEST_INSTANCE_UID_1],
            'vr': 'UI',
        },
    }]
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.QidoRs.return_value = all_instances_json
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    instances = dwi.get_instances(self._study_path_1)
    self.assertEqual(
        instances[0],
        dicom_web_interface.DicomObject(
            self._instance_path_1, all_instances_json[0]
        ),
        (
            'All DICOM objects returned should have a type of INSTANCE, with '
            'corresponding DICOM path and original DICOM tags attached.'
        ),
    )

  def test_get_instances_with_series_level_path(self):
    all_instances_json = [{
        '00080018': {
            'Value': [dicom_test_utils.TEST_INSTANCE_UID_1],
            'vr': 'UI',
        },
        '00280010': {'Value': [500], 'vr': 'US'},
        '00280002': {'Value': [3], 'vr': 'US'},
        '00280100': {'Value': [8], 'vr': 'US'},
        '00280102': {'Value': [7], 'vr': 'US'},
        '00280011': {'Value': [500], 'vr': 'US'},
        '00480006': {'Value': [98816], 'vr': 'UL'},
        '00480007': {'Value': [199168], 'vr': 'UL'},
    }]
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.QidoRs.return_value = all_instances_json
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    instances = dwi.get_instances(self._series_path_1)
    self.assertEqual(
        instances[0],
        dicom_web_interface.DicomObject(
            self._instance_path_1, all_instances_json[0]
        ),
        (
            'The DICOM object returned should have a type of INSTANCE, with '
            'corresponding DICOM path and original DICOM tags attached.'
        ),
    )
    self.assertEqual(500, instances[0].get_value(tags.ROWS))
    self.assertEqual(500, instances[0].get_value(tags.COLUMNS))
    self.assertEqual(3, instances[0].get_value(tags.SAMPLES_PER_PIXEL))
    self.assertEqual(8, instances[0].get_value(tags.BITS_ALLOCATED))
    self.assertEqual(7, instances[0].get_value(tags.HIGH_BIT))
    self.assertEqual(
        98816, instances[0].get_value(tags.TOTAL_PIXEL_MATRIX_COLUMNS)
    )
    self.assertEqual(
        199168, instances[0].get_value(tags.TOTAL_PIXEL_MATRIX_ROWS)
    )

  def test_get_frame_image_with_invalid_input_raise_error(self):
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    with self.subTest(name='Store'):
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_frame_image(
            self._dicom_store_path,
            1,
            dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN,
        )
    with self.subTest(name='Study'):
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_frame_image(
            self._study_path_1,
            1,
            dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN,
        )
    with self.subTest(name='Series'):
      with self.assertRaises(ez_wsi_errors.DicomPathError):
        dwi.get_frame_image(
            self._series_path_1,
            1,
            dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN,
        )

  def test_get_frame_image_with_valid_input(self):
    expected_image = b'\x01\x02\x03\x04'
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    mock_dicom_web_client.WadoRs.return_value = expected_image
    dwi = dicom_web_interface.DicomWebInterface(mock_dicom_web_client)
    self.assertEqual(
        expected_image,
        dwi.get_frame_image(
            self._instance_path_1,
            1,
            dicom_web_interface.TranscodeDicomFrame.UNCOMPRESSED_LITTLE_ENDIAN,
        ),
    )

  def test_constructor_base_url_override(self):
    fake_base_url = 'fake_base_url'
    mock_dicom_web_client = mock.create_autospec(dicom_web.DicomWebClient)
    dwi = dicom_web_interface.DicomWebInterface(
        mock_dicom_web_client, fake_base_url
    )
    self.assertEqual(dwi._dicom_web_base_url, fake_base_url)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.dicom_frame_decoder."""
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import cv2
from ez_wsi_dicomweb import dicom_frame_decoder
from ez_wsi_dicomweb import dicom_test_utils
import numpy as np


def _read_test_jpg() -> bytes:
  with open(f'{dicom_test_utils.TEST_DATA_PATH}/google.jpg', 'rb') as infile:
    return infile.read()


def _expected_np_array() -> np.ndarray:
  return np.load(
      f'{dicom_test_utils.TEST_DATA_PATH}/google.npy', allow_pickle=False
  )


class DicomFrameDecoderTest(parameterized.TestCase):

  @parameterized.parameters([
      ('1.2.840.10008.1.2.4.50', True),
      ('1.2.840.10008.1.2.4.90', True),
      ('1.2.840.10008.1.2.4.91', True),
      ('1.2.840.10008.1.2', False),
      ('', False),
      ('1.2.3', False),
  ])
  def test_can_decompress_dicom_transfer_syntax(
      self, transfer_syntax, can_decode
  ):
    self.assertEqual(
        dicom_frame_decoder.can_decompress_dicom_transfer_syntax(
            transfer_syntax
        ),
        can_decode,
    )

  @mock.patch.object(cv2, 'imdecode')
  def test_decode_compressed_frame_bytes_pil(self, mock_imdecode):
    mock_imdecode.return_value = None
    img = _read_test_jpg()
    self.assertTrue(dicom_frame_decoder._PIL_LOADED)
    decoded_img = dicom_frame_decoder.decode_dicom_compressed_frame_bytes(img)
    self.assertTrue(np.array_equal(decoded_img, _expected_np_array()))

  def test_decode_compressed_frame_bytes_cv2(self):
    pil_loaded = dicom_frame_decoder._PIL_LOADED
    img = _read_test_jpg()
    try:
      dicom_frame_decoder._PIL_LOADED = False
      decoded_img = dicom_frame_decoder.decode_dicom_compressed_frame_bytes(img)
      self.assertTrue(np.array_equal(decoded_img, _expected_np_array()))
    finally:
      dicom_frame_decoder._PIL_LOADED = pil_loaded


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ez_wsi_dicomweb.dicom_store."""
from absl.testing import absltest
from absl.testing import parameterized

from ez_wsi_dicomweb import dicom_store
from ez_wsi_dicomweb import dicom_test_utils


class DicomStoreTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # Load DICOM objects for the testing slide.
    self.mock_dwi = dicom_test_utils.create_mock_dicom_web_interface(
        dicom_test_utils.SAMPLE_INSTANCES_PATH
    )

  def test_get_slide(self):
    slide = dicom_store.DicomStore(
        dicom_test_utils.TEST_STORE_PATH, True, self.mock_dwi
    ).get_slide(dicom_test_utils.TEST_SLIDE_UID_1)
    self.assertEqual('project_name', slide.path.project_id)
    self.assertEqual('us-west1', slide.path.location)
    self.assertEqual(
        dicom_test_utils.TEST_SLIDE_UID_1.split(':')[0], slide.path.study_uid
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for pixel_spacing_converter."""
from absl.testing import absltest
from absl.testing import parameterized
from ez_wsi_dicomweb import ez_wsi_errors
from ez_wsi_dicomweb import magnification as mag_lib
from ez_wsi_dicomweb import pixel_spacing_converter


class PixelSpacingConverterTest(parameterized.TestCase):

  def test_magnification_to_pixel_spacing(self):
    magnification = mag_lib.Magnification.FromString('5X')
    ps = pixel_spacing_converter.magnification_to_pixel_spacing(magnification)

    self.assertEqual(0.002, ps.pixel_spacing_mm)

  def test_magnification_string_to_pixel_spacing(self):
    ps = pixel_spacing_converter.magnification_string_to_pixel_spacing('5X')

    self.assertEqual(0.002, ps.pixel_spacing_mm)

  def test_magnification_to_pixel_spacing_raises(self):
    magnification_unknown = mag_lib.Magnification.Unknown()
    with self.assertRaises(ez_wsi_errors.MagnificationLevelNotFoundError):
      pixel_spacing_converter.magnification_to_pixel_spacing(
          magnification_unknown
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""DICOMWeb API Interface."""
import dataclasses
import enum
import posixpath
from typing import Any, Collection, Dict, List, Optional, Sequence
import urllib

from absl import logging
from ez_wsi_dicomweb import ez_wsi_errors

from hcls_imaging_ml_toolkit import dicom_json
from hcls_imaging_ml_toolkit import dicom_path
from hcls_imaging_ml_toolkit import dicom_web
from hcls_imaging_ml_toolkit import tags

# The default Google HealthCare DICOMWeb API entry.
_DEFAULT_DICOMWEB_BASE_URL = 'https://healthcare.googleapis.com/v1'
# The default DICOM tags to retrieve for an instance.
_DEFAULT_INSTANCE_TAGS = (
    tags.SOP_INSTANCE_UID,
    tags.INSTANCE_NUMBER,
    tags.ROWS,
    tags.COLUMNS,
    tags.SAMPLES_PER_PIXEL,
    tags.BITS_ALLOCATED,
    tags.HIGH_BIT,
    tags.PIXEL_REPRESENTATION,
    tags.PIXEL_SPACING,
    tags.NUMBER_OF_FRAMES,
    tags.FRAME_INCREMENT_POINTER,
    tags.IMAGE_VOLUME_WIDTH,
    tags.IMAGE_VOLUME_HEIGHT,
    tags.IMAGE_VOLUME_DEPTH,
    tags.TOTAL_PIXEL_MATRIX_COLUMNS,
    tags.TOTAL_PIXEL_MATRIX_ROWS,
    tags.TOTAL_PIXEL_MATRIX_ORIGIN_SEQUENCE,
    tags.SPECIMEN_LABEL_IN_IMAGE,
    tags.EXTENDED_DEPTH_OF_FIELD,
    tags.LABEL_TEXT,
    tags.BARCODE_VALUE,
    tags.CONCATENATION_FRAME_OFFSET_NUMBER,
    tags.IMAGE_TYPE,
    tags.TRANSFER_SYNTAX_UID,
)


# Transfer syntax defining how DICOM frames should be transcoded.
# https://cloud.google.com/healthcare-api/docs/dicom#dicom_frames
class TranscodeDicomFrame(enum.Enum):
  DO_NOT_TRANSCODE = '*'
  UNCOMPRESSED_LITTLE_ENDIAN = '1.2.840.10008.1.2.1'


@dataclasses.dataclass(frozen=True)
class DicomObject:
  """Represents a DICOM object by its path and DICOM tags.

  Attributes:
    path: The DICOM path to the object in a DICOMStore.
    dicom_tags: A list of DICOM tags that are associated with the DICOM object.
  """

  path: dicom_path.Path
  dicom_tags: Dict[str, Any]

  def __post_init__(self):
    """Post __init__.

    Raises:
      DicomPathError if self.path type is not STUDY, SERIES or
      INSTANCE.
    """
    if self.path.type not in (
        dicom_path.Type.STUDY,
        dicom_path.Type.SERIES,
        dicom_path.Type.INSTANCE,
    ):
      raise ez_wsi_errors.DicomPathError(
          f'The DICOM path has an invalid type: {self.path.type}'
      )

  def type(self) -> dicom_path.Type:
    """Returns the type of the DICOM  path."""
    return self.path.type

  def get_value(self, tag: tags.DicomTag) -> Any:
    """Returns the first value for the tag in dicom_tags.

    For many DICOM tags this is going to be the only value in the list. If you
    wish to fetch the entire list of tags instead, use get_list_value().

    Args:
      tag: The tag to lookup values for.

    Returns:
      The first value from dicom_tags corresponding to the tag or None if the
      tag is not present in dicom_tags.
    """
    return dicom_json.GetValue(self.dicom_tags, tag)

  def get_list_value(self, tag: tags.DicomTag) -> Optional[List[Any]]:
    """Returns the value list for the tag from dicom_tags.

    Args:
      tag: The tag to lookup values for.

    Returns:
      The value list corresponding to the tag or None if the tag is not present
      in dicom_tags.
    """
    return dicom_json.GetList(self.dicom_tags, tag)


class DicomWebInterface(object):
  """A Python interface of the DICOMWeb API.

  It wraps around the HealthCare DICOMWeb API to fetch DICOM objects, such as
  studies, series, instances, and dicom_tags associated with those objects.
  """

  def __init__(
      self,
      dicom_web_client: dicom_web.DicomWebClient,
      dicom_web_base_url: str = _DEFAULT_DICOMWEB_BASE_URL,
  ):
    """Constructor.

    Args:
      dicom_web_client: Configured DicomWeb client.
      dicom_web_base_url: Optional parameter that is used as the base of the url
        for all REST calls. Defaults to the Google HealthCare DICOMWeb API base.
    """
    self._dicom_web_client = dicom_web_client
    self._dicom_web_base_url = dicom_web_base_url

  def _make_api_url(self, base_path: dicom_path.Path, api_entry) -> str:
    """Constructs a full http url for the input base path and api entry.

    Args:
      base_path: The base path of the resource upon which the api will be
        called. This path should take the form of:
        'projects/{}/locations/{}/datasets/{}/dicomStores/{}'
      api_entry: The api path to attach after the base path.

    Returns:
      The full http url for calling the API.
    """
    if base_path.type == dicom_path.Type.STORE:
      base_path_str = base_path.dicomweb_path_str
    else:
      base_path_str = str(base_path)
    return posixpath.join(self._dicom_web_base_url, base_path_str, api_entry)

  def get_studies(
      self, dicom_store_path: dicom_path.Path
  ) -> Sequence[DicomObject]:
    """Gets all study objects from the input DICOMStore.

    Args:
      dicom_store_path: The path to the DICOMStore to fetch studies from.

    Returns:
      A sequence of the DICOM study objects.

    Raises:
      DicomPathError: If the input path is not a valid store-level
      dicom path.
    """
    if dicom_store_path.type != dicom_path.Type.STORE:
      raise ez_wsi_errors.DicomPathError(
          'A store-level path is required.'
          f'Found: {dicom_store_path.type} path found in {dicom_store_path}'
      )
    api_url = self._make_api_url(dicom_store_path, 'studies')
    json_results = self._dicom_web_client.QidoRs(api_url)
    if not json_results:
      logging.warn('No studies on the requested %s', api_url)
      return []

    return [
        _build_dicom_object(dicom_path.Type.STUDY, dicom_store_path, dicom_tags)
        for dicom_tags in json_results
    ]

  def get_series(
      self,
      parent_path: dicom_path.Path,
      dicom_tags: Optional[Dict[str, Any]] = None,
  ) -> Sequence[DicomObject]:
    """Gets all series objects under the input parent path.

    Args:
      parent_path: The path to a DICOMStore or a study to fetch series from.
      dicom_tags: Tags to search for, if provided.

    Returns:
      A sequence of the series objects.

    Raises:
      DicomPathError: If the input path is not a valid store-level or
      study-level
      DICOM path.
    """
    if parent_path.type not in (dicom_path.Type.STORE, dicom_path.Type.STUDY):
      raise ez_wsi_errors.DicomPathError(
          'A store-level or study-level path is required.'
          f'Found: {parent_path.type} path found in {parent_path}'
      )
    series_query = (
        f'series?{urllib.parse.urlencode(dicom_tags)}'
        if dicom_tags
        else 'series'
    )

    api_url = self._make_api_url(parent_path, series_query)
    json_results = self._dicom_web_client.QidoRs(api_url)
    if not json_results:
      logging.warn('No series on the requested %s', api_url)
      return []

    return [
        _build_dicom_object(dicom_path.Type.SERIES, parent_path, dicom_tags)
        for dicom_tags in json_results
    ]

  def get_instances(
      self, parent_path: dicom_path.Path
  ) -> Sequence[DicomObject]:
    """Gets all instances under the input parent path.

    Args:
      parent_path: The path to a DICOMStore, a study or a series.

    Returns:
      A sequence of the DICOM instance objects.

    Raises:
      DicomPathError: If the input path is not a valid store-level,
      study-level or series-level DICOM path.
    """
    if parent_path.type not in (
        dicom_path.Type.STORE,
        dicom_path.Type.STUDY,
        dicom_path.Type.SERIES,
    ):
      raise ez_wsi_errors.DicomPathError(
          'A store-level, study-level or series-level path is required.'
          f'Found: {parent_path.type} path found in {parent_path}'
      )
    api_url = self._make_api_url(
        parent_path,
        f'instances/?{_get_qido_suffix(_DEFAULT_INSTANCE_TAGS)}',
    )
    json_results = self._dicom_web_client.QidoRs(api_url)
    if not json_results:
      logging.warn('No instances on the requested %s', api_url)
      return []

    return [
        _build_dicom_object(dicom_path.Type.INSTANCE, parent_path, dicom_tags)
        for dicom_tags in json_results
    ]

  def get_frame_image(
      self,
      instance_path: dicom_path.Path,
      frame_index: int,
      transcode_frame: TranscodeDicomFrame,
  ) -> bytes:
    """Retrieves the BGR image pixels of a frame from an instance.

    Args:
      instance_path: The path to a DICOM instance.
      frame_index: The index of the DICOM frame within the requested instance.
      transcode_frame: How DICOM frame should be transcoded.

    Returns:
      The image buffer that contains the pixels of the frame.

    Raises:
      DicomPathError if the input path is not an instance.
    """
    if instance_path.type != dicom_path.Type.INSTANCE:
      raise ez_wsi_errors.DicomPathError(
          'An instance-level path is required. '
          f'Found: {instance_path.type} path found in {instance_path}'
      )
    api_url = self._make_api_url(instance_path, f'frames/{frame_index}')
    return self._dicom_web_client.WadoRs(
        api_url,
        (
            'multipart/related; type=application/octet-stream; '
            f'transfer-syntax={transcode_frame.value}'
        ),
    )


def _build_dicom_object(
    object_type: dicom_path.Type,
    parent_path: dicom_path.Path,
    dicom_tags: Dict[str, Any],
) -> DicomObject:
  """Creates a DICOM object from a DICOM parent path and a set of tags.

  Args:
    object_type: The type of the object to construct.
    parent_path: the DICOM path pointing to the parent of the DICOM object to be
      constructed.
    dicom_tags: DICOM tags associated for the object construction.

  Returns:
    An instance of DicomObject constructed with proper type.
  Raises:
    DicomPathError if the type of the path constructed for the DICOM
    object does not match the input object_type.
  """
  path = dicom_path.FromPath(
      parent_path,
      # the study_uid is None for a store-level path
      study_uid=dicom_json.GetValue(dicom_tags, tags.STUDY_INSTANCE_UID),
      # the series_uid is None for a store-level or tudy-level path
      series_uid=dicom_json.GetValue(dicom_tags, tags.SERIES_INSTANCE_UID),
      instance_uid=dicom_json.GetValue(dicom_tags, tags.SOP_INSTANCE_UID),
  )
  if path.type != object_type:
    raise ez_wsi_errors.DicomPathError(
        f'The parent path and DICOM tags do not yield a path({path.type}) as '
        f'expected: {object_type}'
    )
  return DicomObject(path, dicom_tags)


def _get_qido_suffix(tags_to_fetch: Collection[tags.DicomTag]) -> str:
  """Returns the suffix to be used in a QIDO-RS query.

  Args:
    tags_to_fetch: DICOM tags to be fetched for each instance.
  """
  return urllib.parse.urlencode(
      {'includefield': [tag.number for tag in tags_to_fetch]}, doseq=True
  )
